## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import argparse
import os
import os.path
import platform
import readline
import sys
import time

from binascii import hexlify
from binascii import unhexlify
from select import select
from threading import Lock
from threading import Thread

if platform.system() == "Linux":
    import hidapi.udev_impl as hidapi
else:
    import hidapi.hidapi_impl as hidapi

LOGITECH_VENDOR_ID = 0x046D

interactive = os.isatty(0)
prompt = "?? Input: " if interactive else ""
start_time = time.time()


def strhex(d):
    return hexlify(d).decode("ascii").upper()


print_lock = Lock()


def _print(marker, data, scroll=False):
    t = time.time() - start_time
    if isinstance(data, str):
        s = f"{marker} {data}"
    else:
        hexs = strhex(data)
        s = "%s (% 8.3f) [%s %s %s %s] %s" % (marker, t, hexs[0:2], hexs[2:4], hexs[4:8], hexs[8:], repr(data))

    with print_lock:
        # allow only one thread at a time to write to the console, otherwise
        # the output gets garbled, especially with ANSI codes.

        if interactive and scroll:
            # scroll the entire screen above the current line up by 1 line
            sys.stdout.write(
                "\033[s"  # save cursor position
                "\033[S"  # scroll up
                "\033[A"  # cursor up
                "\033[L"  # insert 1 line
                "\033[G"
            )  # move cursor to column 1
        sys.stdout.write(s)
        if interactive and scroll:
            # restore cursor position
            sys.stdout.write("\033[u")
        else:
            sys.stdout.write("\n")

        # flush stdout manually...
        # because trying to open stdin/out unbuffered programmatically
        # works much too differently in Python 2/3
        sys.stdout.flush()


def _error(text, scroll=False):
    _print("!!", text, scroll)


def _continuous_read(handle, timeout=2000):
    while True:
        try:
            reply = hidapi.read(handle, 128, timeout)
        except OSError as e:
            _error(f"Read failed, aborting: {str(e)}", True)
            break
        assert reply is not None
        if reply:
            _print(">>", reply, True)


def _validate_input(line, hidpp=False):
    try:
        data = unhexlify(line.encode("ascii"))
    except Exception as e:
        _error(f"Invalid input: {str(e)}")
        return None

    if hidpp:
        if len(data) < 4:
            _error("Invalid HID++ request: need at least 4 bytes")
            return None
        if data[:1] not in b"\x10\x11":
            _error("Invalid HID++ request: first byte must be 0x10 or 0x11")
            return None
        if data[1:2] not in b"\xff\x00\x01\x02\x03\x04\x05\x06\x07":
            _error("Invalid HID++ request: second byte must be 0xFF or one of 0x00..0x07")
            return None
        if data[:1] == b"\x10":
            if len(data) > 7:
                _error("Invalid HID++ request: maximum length of a 0x10 request is 7 bytes")
                return None
            while len(data) < 7:
                data = (data + b"\x00" * 7)[:7]
        elif data[:1] == b"\x11":
            if len(data) > 20:
                _error("Invalid HID++ request: maximum length of a 0x11 request is 20 bytes")
                return None
            while len(data) < 20:
                data = (data + b"\x00" * 20)[:20]

    return data


def _open(args):
    def matchfn(bid, vid, pid, _a, _b):
        if vid == LOGITECH_VENDOR_ID:
            return {"vid": vid}

    device = args.path
    d = None
    if not device:
        for d in hidapi.enumerate(matchfn):
            if (d.hidpp_short or d.hidpp_long) and (args.id is None or args.id.lower() == d.product_id.lower()):
                device = d.path
                break
        if not device:
            sys.exit("!! No HID++ receiver found.")
    if not device:
        sys.exit("!! Device path required.")

    handle = hidapi.open_path(device)
    if not handle:
        sys.exit(f"!! Failed to open {device}, aborting.")
    print(
        ".. Opened device %r, vendor %r product %r serial %r."
        % (
            device,
            hidapi.get_manufacturer(handle) or d.vendor_id if d else None,
            hidapi.get_product(handle) or d.product_id if d else None,
            hidapi.get_serial(handle),
        )
    )
    if args.hidpp:
        if hidapi.get_manufacturer(handle) is not None and hidapi.get_manufacturer(handle) != b"Logitech":
            sys.exit("!! Only Logitech devices support the HID++ protocol.")
        print(".. HID++ validation enabled.")
    else:
        if hidapi.get_manufacturer(handle) == b"Logitech" and b"Receiver" in hidapi.get_product(handle):
            args.hidpp = True
            print(".. Logitech receiver detected, HID++ validation enabled.")

    return handle


def _parse_arguments():
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument("--history", help="history file (default ~/.hidconsole-history)")
    arg_parser.add_argument("--hidpp", action="store_true", help="ensure input data is a valid HID++ request")
    arg_parser.add_argument("command", nargs="?", help="command to send (otherwise get commands from input)")
    group = arg_parser.add_mutually_exclusive_group()
    group.add_argument("-p", "--path", help="HID raw device to connect to (/dev/hidrawX); ")
    group.add_argument("-i", "--id", help="product ID of device to connect to (XXXX)")
    return arg_parser.parse_args()


def main():
    args = _parse_arguments()
    handle = _open(args)

    if args.command:  # send a command
        data = _validate_input(args.command, args.hidpp)
        if data:
            hidapi.write(handle, data)
            reply = hidapi.read(handle, 128, 1)
            if reply:
                hexs = strhex(reply)
                s = "[%s %s %s %s] %s" % (hexs[0:2], hexs[2:4], hexs[4:8], hexs[8:], repr(data))
                print(s)
        exit()

    if interactive:
        print(".. Press ^C/^D to exit, or type hex bytes to write to the device.")

        if args.history is None:
            args.history = os.path.join(os.path.expanduser("~"), ".hidconsole-history")
        try:
            readline.read_history_file(args.history)
        except Exception:
            # file may not exist yet
            pass

    try:
        t = Thread(target=_continuous_read, args=(handle,))
        t.daemon = True
        t.start()

        if interactive:
            # move the cursor at the bottom of the screen
            sys.stdout.write("\033[300B")  # move cusor at most 300 lines down, don't scroll

        while t.is_alive():
            line = input(prompt)
            line = line.strip().replace(" ", "")
            # print ("line", line)
            if not line:
                continue

            data = _validate_input(line, args.hidpp)
            if data is None:
                continue

            _print("<<", data)
            hidapi.write(handle, data)
            # wait for some kind of reply
            if args.hidpp and not interactive:
                rlist, wlist, xlist = select([handle], [], [], 1)
                if data[1:2] == b"\xff":
                    # the receiver will reply very fast, in a few milliseconds
                    time.sleep(0.010)
                else:
                    # the devices might reply quite slow
                    time.sleep(0.700)
    except EOFError:
        if interactive:
            print("")
        else:
            time.sleep(1)

    finally:
        hidapi.close(handle)
        if interactive:
            readline.write_history_file(args.history)


if __name__ == "__main__":
    main()



## Copyright (C) 2012-2013  Daniel Pavel
## Copyright (C) 2014-2024  Solaar Contributors https://pwr-solaar.github.io/Solaar/
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
from __future__ import annotations

from enum import Flag
from enum import IntEnum
from typing import List

from .common import NamedInts

"""HID constants for HID++ 1.0.

Most of them as defined by the official Logitech HID++ 1.0
documentation, some of them guessed.
"""

DEVICE_KIND = NamedInts(
    unknown=0x00,
    keyboard=0x01,
    mouse=0x02,
    numpad=0x03,
    presenter=0x04,
    remote=0x07,
    trackball=0x08,
    touchpad=0x09,
    tablet=0x0A,
    gamepad=0x0B,
    joystick=0x0C,
    headset=0x0D,  # not from Logitech documentation
    remote_control=0x0E,  # for compatibility with HID++ 2.0
    receiver=0x0F,  # for compatibility with HID++ 2.0
)


class PowerSwitchLocation(IntEnum):
    UNKNOWN = 0x00
    BASE = 0x01
    TOP_CASE = 0x02
    EDGE_OF_TOP_RIGHT_CORNER = 0x03
    TOP_LEFT_CORNER = 0x05
    BOTTOM_LEFT_CORNER = 0x06
    TOP_RIGHT_CORNER = 0x07
    BOTTOM_RIGHT_CORNER = 0x08
    TOP_EDGE = 0x09
    RIGHT_EDGE = 0x0A
    LEFT_EDGE = 0x0B
    BOTTOM_EDGE = 0x0C

    @classmethod
    def location(cls, loc: int) -> PowerSwitchLocation:
        try:
            return cls(loc)
        except ValueError:
            return cls.UNKNOWN


class NotificationFlag(Flag):
    """Some flags are used both by devices and receivers.

    The Logitech documentation mentions that the first and last (third)
    byte are used for devices while the second is used for the receiver.
    In practise, the second byte is also used for some device-specific
    notifications (keyboard illumination level). Do not simply set all
    notification bits if the software does not support it. For example,
    enabling keyboard_sleep_raw makes the Sleep key a no-operation
    unless the software is updated to handle that event.

    Observations:
    - wireless and software present seen on receivers,
    reserved_r1b4 as well
    - the rest work only on devices as far as we can tell right now
    In the future would be useful to have separate enums for receiver
    and device notification flags, but right now we don't know enough.
    Additional flags taken from https://drive.google.com/file/d/0BxbRzx7vEV7eNDBheWY0UHM5dEU/view?usp=sharing
    """

    @classmethod
    def flag_names(cls, flags) -> List[str]:
        """Extract the names of the flags from the integer."""
        return flags.name.replace("_", " ").lower().split("|")

    NUMPAD_NUMERICAL_KEYS = 0x800000
    F_LOCK_STATUS = 0x400000
    ROLLER_H = 0x200000
    BATTERY_STATUS = 0x100000  # send battery charge notifications (0x07 or 0x0D)
    MOUSE_EXTRA_BUTTONS = 0x080000
    ROLLER_V = 0x040000
    POWER_KEYS = 0x020000  # system control keys such as Sleep
    KEYBOARD_MULTIMEDIA_RAW = 0x010000  # consumer controls such as Mute and Calculator
    MULTI_TOUCH = 0x001000  # notify on multi-touch changes
    SOFTWARE_PRESENT = 0x000800  # software is controlling part of device behaviour
    LINK_QUALITY = 0x000400  # notify on link quality changes
    UI = 0x000200  # notify on UI changes
    WIRELESS = 0x000100  # notify when the device wireless goes on/off-line
    CONFIGURATION_COMPLETE = 0x000004
    VOIP_TELEPHONY = 0x000002
    THREED_GESTURE = 0x000001


def flags_to_str(flags, fallback: str) -> str:
    flag_names = []
    if flags is not None and flags is not False:
        if flags.value == 0:
            flag_names = (fallback,)
        else:
            flag_names = NotificationFlag.flag_names(flags)
    return f"\n{' ':15}".join(sorted(flag_names))


class ErrorCode(IntEnum):
    INVALID_SUB_ID_COMMAND = 0x01
    INVALID_ADDRESS = 0x02
    INVALID_VALUE = 0x03
    CONNECTION_REQUEST_FAILED = 0x04
    TOO_MANY_DEVICES = 0x05
    ALREADY_EXISTS = 0x06
    BUSY = 0x07
    UNKNOWN_DEVICE = 0x08
    RESOURCE_ERROR = 0x09
    REQUEST_UNAVAILABLE = 0x0A
    UNSUPPORTED_PARAMETER_VALUE = 0x0B
    WRONG_PIN_CODE = 0x0C


class PairingError(IntEnum):
    DEVICE_TIMEOUT = 0x01
    DEVICE_NOT_SUPPORTED = 0x02
    TOO_MANY_DEVICES = 0x03
    SEQUENCE_TIMEOUT = 0x06

    @property
    def label(self) -> str:
        return self.name.lower().replace("_", " ")


class BoltPairingError(IntEnum):
    DEVICE_TIMEOUT = 0x01
    FAILED = 0x02

    @property
    def label(self) -> str:
        return self.name.lower().replace("_", " ")


class Registers(IntEnum):
    """Known HID registers.

    Devices usually have a (small) sub-set of these. Some registers are only
    applicable to certain device kinds (e.g. smooth_scroll only applies to mice).
    """

    # Generally applicable
    NOTIFICATIONS = 0x00
    FIRMWARE = 0xF1

    # only apply to receivers
    RECEIVER_CONNECTION = 0x02
    RECEIVER_PAIRING = 0xB2
    DEVICES_ACTIVITY = 0x2B3
    RECEIVER_INFO = 0x2B5
    BOLT_DEVICE_DISCOVERY = 0xC0
    BOLT_PAIRING = 0x2C1
    BOLT_UNIQUE_ID = 0x02FB

    # only apply to devices
    MOUSE_BUTTON_FLAGS = 0x01
    KEYBOARD_HAND_DETECTION = 0x01
    DEVICES_CONFIGURATION = 0x03
    BATTERY_STATUS = 0x07
    KEYBOARD_FN_SWAP = 0x09
    BATTERY_CHARGE = 0x0D
    KEYBOARD_ILLUMINATION = 0x17
    THREE_LEDS = 0x51
    MOUSE_DPI = 0x63

    # notifications
    PASSKEY_REQUEST_NOTIFICATION = 0x4D
    PASSKEY_PRESSED_NOTIFICATION = 0x4E
    DEVICE_DISCOVERY_NOTIFICATION = 0x4F
    DISCOVERY_STATUS_NOTIFICATION = 0x53
    PAIRING_STATUS_NOTIFICATION = 0x54


# Subregisters for receiver_info register
class InfoSubRegisters(IntEnum):
    SERIAL_NUMBER = 0x01  # not found on many receivers
    FW_VERSION = 0x02
    RECEIVER_INFORMATION = 0x03
    PAIRING_INFORMATION = 0x20  # 0x2N, by connected device
    EXTENDED_PAIRING_INFORMATION = 0x30  # 0x3N, by connected device
    DEVICE_NAME = 0x40  # 0x4N, by connected device
    BOLT_PAIRING_INFORMATION = 0x50  # 0x5N, by connected device
    BOLT_DEVICE_NAME = 0x60  # 0x6N01, by connected device


class DeviceFeature(Flag):
    """Features for devices.

    Flags taken from
    https://drive.google.com/file/d/0BxbRzx7vEV7eNDBheWY0UHM5dEU/view?usp=sharing
    """

    @classmethod
    def flag_names(cls, flag_bits: int) -> List[str]:
        """Extract the names of the flags from the integer."""
        indexed = {item.value: item.name for item in cls}

        flag_names = []
        unknown_bits = flag_bits
        for k in indexed:
            # Ensure that the key (flag value) is a power of 2 (a single bit flag)
            assert bin(k).count("1") == 1
            if k & flag_bits == k:
                unknown_bits &= ~k
                flag_names.append(indexed[k].replace("_", " ").lower())

        # Yield any remaining unknown bits
        if unknown_bits != 0:
            flag_names.append(f"unknown:{unknown_bits:06X}")
        return flag_names

    RESERVED1 = 0x010000
    SPECIAL_BUTTONS = 0x020000
    ENHANCED_KEY_USAGE = 0x040000
    FAST_FW_REV = 0x080000
    RESERVED2 = 0x100000
    RESERVED3 = 0x200000
    SCROLL_ACCEL = 0x400000
    BUTTONS_CONTROL_RESOLUTION = 0x800000
    INHIBIT_LOCK_KEY_SOUND = 0x000001
    RESERVED4 = 0x000002
    MX_AIR_3D_ENGINE = 0x000004
    HOST_CONTROL_LEDS = 0x000008
    RESERVED5 = 0x000010
    RESERVED6 = 0x000020
    RESERVED7 = 0x000040
    RESERVED8 = 0x000080



## Copyright (C) 2012-2013  Daniel Pavel
## Copyright (C) 2014-2024  Solaar Contributors https://pwr-solaar.github.io/Solaar/
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""Handles incoming events from the receiver/devices, updating the
object as appropriate.
"""

from __future__ import annotations

import logging
import struct
import threading
import typing

from solaar.i18n import _

from . import base
from . import common
from . import diversion
from . import hidpp10
from . import hidpp10_constants
from . import hidpp20
from . import settings_templates
from .common import Alert
from .common import BatteryStatus
from .common import Notification
from .hidpp10_constants import Registers
from .hidpp20_constants import SupportedFeature

if typing.TYPE_CHECKING:
    from .base import HIDPPNotification
    from .device import Device
    from .receiver import Receiver

logger = logging.getLogger(__name__)

NotificationHandler = typing.Callable[["Receiver", "HIDPPNotification"], bool]

_hidpp10 = hidpp10.Hidpp10()
_hidpp20 = hidpp20.Hidpp20()

notification_lock = threading.Lock()


def process(device: Device | Receiver, notification: HIDPPNotification):
    """Handle incoming events (notification) from device or receiver."""
    assert device
    assert notification

    if not device.isDevice:
        return process_receiver_notification(device, notification)
    return process_device_notification(device, notification)


def process_receiver_notification(receiver: Receiver, notification: HIDPPNotification) -> bool | None:
    """Process event messages from receivers."""
    event_handler_mapping: dict[int, NotificationHandler] = {
        Notification.PAIRING_LOCK: handle_pairing_lock,
        Registers.DEVICE_DISCOVERY_NOTIFICATION: handle_device_discovery,
        Registers.DISCOVERY_STATUS_NOTIFICATION: handle_discovery_status,
        Registers.PAIRING_STATUS_NOTIFICATION: handle_pairing_status,
        Registers.PASSKEY_PRESSED_NOTIFICATION: handle_passkey_pressed,
        Registers.PASSKEY_REQUEST_NOTIFICATION: handle_passkey_request,
    }

    try:
        handler_func = event_handler_mapping[notification.sub_id]
        return handler_func(receiver, notification)
    except KeyError:
        pass

    assert notification.sub_id in [
        Notification.CONNECT_DISCONNECT,
        Notification.DJ_PAIRING,
        Notification.CONNECTED,
        Notification.RAW_INPUT,
        Notification.POWER,
    ]

    logger.warning(f"{receiver}: unhandled notification {notification}")


def process_device_notification(device: Device, notification: HIDPPNotification):
    """Process event messages from devices."""

    # incoming packets with SubId >= 0x80 are supposedly replies from HID++ 1.0 requests, should never get here
    assert notification.sub_id & 0x80 == 0

    if notification.sub_id == Notification.NO_OPERATION:
        # dispose it
        return False

    # Allow the device object to handle the notification using custom per-device state.
    handling_ret = device.handle_notification(notification)
    if handling_ret is not None:
        return handling_ret

    # 0x40 to 0x7F appear to be HID++ 1.0 or DJ notifications
    if notification.sub_id >= 0x40:
        if notification.report_id == base.DJ_MESSAGE_ID:
            return _process_dj_notification(device, notification)
        else:
            return _process_hidpp10_notification(device, notification)

    # These notifications are from the device itself, so it must be active
    device.online = True
    # At this point, we need to know the device's protocol, otherwise it's possible to not know how to handle it.
    assert device.protocol is not None

    # some custom battery events for HID++ 1.0 devices
    if device.protocol < 2.0:
        return _process_hidpp10_custom_notification(device, notification)

    # assuming 0x00 to 0x3F are feature (HID++ 2.0) notifications
    if not device.features:
        logger.warning("%s: feature notification but features not set up: %02X %s", device, notification.sub_id, notification)
        return False

    return _process_feature_notification(device, notification)


def _process_dj_notification(device: Device, notification: HIDPPNotification):
    if logger.isEnabledFor(logging.DEBUG):
        logger.debug("%s (%s) DJ %s", device, device.protocol, notification)

    if notification.sub_id == Notification.CONNECT_DISCONNECT:
        # do all DJ paired notifications also show up as HID++ 1.0 notifications?
        if logger.isEnabledFor(logging.INFO):
            logger.info("%s: ignoring DJ unpaired: %s", device, notification)
        return True

    if notification.sub_id == Notification.DJ_PAIRING:
        # do all DJ paired notifications also show up as HID++ 1.0 notifications?
        if logger.isEnabledFor(logging.INFO):
            logger.info("%s: ignoring DJ paired: %s", device, notification)
        return True

    if notification.sub_id == Notification.CONNECTED:
        connected = not notification.address & 0x01
        if logger.isEnabledFor(logging.INFO):
            logger.info("%s: DJ connection: %s %s", device, connected, notification)
        device.changed(active=connected, alert=Alert.NONE, reason=_("connected") if connected else _("disconnected"))
        return True

    logger.warning("%s: unrecognized DJ %s", device, notification)


def _process_hidpp10_custom_notification(device: Device, notification: HIDPPNotification):
    if logger.isEnabledFor(logging.DEBUG):
        logger.debug("%s (%s) custom notification %s", device, device.protocol, notification)

    if notification.sub_id in (Registers.BATTERY_STATUS, Registers.BATTERY_CHARGE):
        assert notification.data[-1:] == b"\x00"
        data = chr(notification.address).encode() + notification.data
        device.set_battery_info(hidpp10.parse_battery_status(notification.sub_id, data))
        return True

    logger.warning("%s: unrecognized %s", device, notification)


def _process_hidpp10_notification(device: Device, notification: HIDPPNotification):
    if notification.sub_id == Notification.CONNECT_DISCONNECT:  # device unpairing
        if notification.address == 0x02:
            # device un-paired
            device.wpid = None
            if device.number in device.receiver:
                del device.receiver[device.number]
            device.changed(active=False, alert=Alert.ALL, reason=_("unpaired"))
        ##            device.status = None
        else:
            logger.warning("%s: disconnection with unknown type %02X: %s", device, notification.address, notification)
        return True

    if notification.sub_id == Notification.DJ_PAIRING:  # device connection (and disconnection)
        flags = ord(notification.data[:1]) & 0xF0
        if notification.address == 0x02:  # very old 27 MHz protocol
            wpid = "00" + common.strhex(notification.data[2:3])
            link_established = True
            link_encrypted = bool(flags & 0x80)
        elif notification.address > 0x00:  # all other protocols are supposed to be almost the same
            wpid = common.strhex(notification.data[2:3] + notification.data[1:2])
            link_established = not (flags & 0x40)
            link_encrypted = bool(flags & 0x20) or notification.address == 0x10  # Bolt protocol always encrypted
        else:
            logger.warning(
                "%s: connection notification with unknown protocol %02X: %s", device.number, notification.address, notification
            )
            return True
        if wpid != device.wpid:
            logger.warning("%s wpid mismatch, got %s", device, wpid)
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(
                "%s: protocol %s connection notification: software=%s, encrypted=%s, link=%s, payload=%s",
                device,
                notification.address,
                bool(flags & 0x10),
                link_encrypted,
                link_established,
                bool(flags & 0x80),
            )
        device.link_encrypted = link_encrypted
        if not link_established and device.receiver:
            hidpp10.set_configuration_pending_flags(device.receiver, 0xFF)
        device.changed(active=link_established)
        return True

    if notification.sub_id == Notification.RAW_INPUT:
        # raw input event? just ignore it
        # if notification.address == 0x01, no idea what it is, but they keep on coming
        # if notification.address == 0x03, appears to be an actual input event, because they only come when input happents
        return True

    if notification.sub_id == Notification.POWER:
        if notification.address == 0x01:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("%s: device powered on", device)
            reason = device.status_string() or _("powered on")
            device.changed(active=True, alert=Alert.NOTIFICATION, reason=reason)
        else:
            logger.warning("%s: unknown %s", device, notification)
        return True

    logger.warning("%s: unrecognized %s", device, notification)


def _process_feature_notification(device: Device, notification: HIDPPNotification):
    old_present, device.present = device.present, True  # the device is generating a feature notification so it must be present
    try:
        feature = device.features.get_feature(notification.sub_id)
    except IndexError:
        logger.warning("%s: notification from invalid feature index %02X: %s", device, notification.sub_id, notification)
        return False

    if logger.isEnabledFor(logging.DEBUG):
        logger.debug(
            "%s: notification for feature %s, report %s, data %s",
            device,
            feature,
            notification.address >> 4,
            common.strhex(notification.data),
        )

    if feature == SupportedFeature.BATTERY_STATUS:
        if notification.address == 0x00:
            device.set_battery_info(hidpp20.decipher_battery_status(notification.data)[1])
        elif notification.address == 0x10:
            if logger.isEnabledFor(logging.INFO):
                logger.info("%s: spurious BATTERY status %s", device, notification)
        else:
            logger.warning("%s: unknown BATTERY %s", device, notification)

    elif feature == SupportedFeature.BATTERY_VOLTAGE:
        if notification.address == 0x00:
            device.set_battery_info(hidpp20.decipher_battery_voltage(notification.data)[1])
        else:
            logger.warning("%s: unknown VOLTAGE %s", device, notification)

    elif feature == SupportedFeature.UNIFIED_BATTERY:
        if notification.address == 0x00:
            device.set_battery_info(hidpp20.decipher_battery_unified(notification.data)[1])
        else:
            logger.warning("%s: unknown UNIFIED BATTERY %s", device, notification)

    elif feature == SupportedFeature.ADC_MEASUREMENT:
        if notification.address == 0x00:
            result = hidpp20.decipher_adc_measurement(notification.data)
            if result:  # if good data and  the device was not present then a push is needed
                device.set_battery_info(result[1])
                device.changed(active=True, alert=Alert.NONE, reason=_("ADC measurement notification"), push=not old_present)
            else:  # this feature is also used to signal device becoming inactive
                device.present = False  # exception to device presence
                device.changed(active=False)
        else:
            logger.warning("%s: unknown ADC MEASUREMENT %s", device, notification)

    elif feature == SupportedFeature.SOLAR_DASHBOARD:
        if notification.data[5:9] == b"GOOD":
            charge, lux, adc = struct.unpack("!BHH", notification.data[:5])
            # guesstimate the battery voltage, emphasis on 'guess'
            # status_text = '%1.2fV' % (adc * 2.67793237653 / 0x0672)
            status_text = BatteryStatus.DISCHARGING
            if notification.address == 0x00:
                device.set_battery_info(common.Battery(charge, None, status_text, None))
            elif notification.address == 0x10:
                if lux > 200:
                    status_text = BatteryStatus.RECHARGING
                device.set_battery_info(common.Battery(charge, None, status_text, None, lux))
            elif notification.address == 0x20:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug("%s: Light Check button pressed", device)
                device.changed(alert=Alert.SHOW_WINDOW)
                # first cancel any reporting
                # device.feature_request(SupportedFeature.SOLAR_DASHBOARD)
                # trigger a new report chain
                reports_count = 15
                reports_period = 2  # seconds
                device.feature_request(SupportedFeature.SOLAR_DASHBOARD, 0x00, reports_count, reports_period)
            else:
                logger.warning("%s: unknown SOLAR CHARGE %s", device, notification)
        else:
            logger.warning("%s: SOLAR CHARGE not GOOD? %s", device, notification)

    elif feature == SupportedFeature.WIRELESS_DEVICE_STATUS:
        if notification.address == 0x00:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("wireless status: %s", notification)
            reason = "powered on" if notification.data[2] == 1 else None
            if notification.data[1] == 1:  # device is asking for software reconfiguration so need to change status
                alert = Alert.NONE
                device.changed(active=True, alert=alert, reason=reason, push=True)
        else:
            logger.warning("%s: unknown WIRELESS %s", device, notification)

    elif feature == SupportedFeature.TOUCHMOUSE_RAW_POINTS:
        if notification.address == 0x00:
            if logger.isEnabledFor(logging.INFO):
                logger.info("%s: TOUCH MOUSE points %s", device, notification)
        elif notification.address == 0x10:
            touch = ord(notification.data[:1])
            button_down = bool(touch & 0x02)
            mouse_lifted = bool(touch & 0x01)
            if logger.isEnabledFor(logging.INFO):
                logger.info("%s: TOUCH MOUSE status: button_down=%s mouse_lifted=%s", device, button_down, mouse_lifted)
        else:
            logger.warning("%s: unknown TOUCH MOUSE %s", device, notification)

    # TODO: what are REPROG_CONTROLS_V{2,3}?
    elif feature == SupportedFeature.REPROG_CONTROLS:
        if notification.address == 0x00:
            if logger.isEnabledFor(logging.INFO):
                logger.info("%s: reprogrammable key: %s", device, notification)
        else:
            logger.warning("%s: unknown REPROG_CONTROLS %s", device, notification)

    elif feature == SupportedFeature.BACKLIGHT2:
        if notification.address == 0x00:
            level = struct.unpack("!B", notification.data[1:2])[0]
            if device.setting_callback:
                device.setting_callback(device, settings_templates.Backlight2Level, [level])

    elif feature == SupportedFeature.REPROG_CONTROLS_V4:
        if notification.address == 0x00:
            if logger.isEnabledFor(logging.DEBUG):
                cid1, cid2, cid3, cid4 = struct.unpack("!HHHH", notification.data[:8])
                logger.debug("%s: diverted controls pressed: 0x%x, 0x%x, 0x%x, 0x%x", device, cid1, cid2, cid3, cid4)
        elif notification.address == 0x10:
            if logger.isEnabledFor(logging.DEBUG):
                dx, dy = struct.unpack("!hh", notification.data[:4])
                logger.debug("%s: rawXY dx=%i dy=%i", device, dx, dy)
        elif notification.address == 0x20:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("%s: received analyticsKeyEvents", device)
        elif logger.isEnabledFor(logging.INFO):
            logger.info("%s: unknown REPROG_CONTROLS_V4 %s", device, notification)

    elif feature == SupportedFeature.HIRES_WHEEL:
        if notification.address == 0x00:
            if logger.isEnabledFor(logging.INFO):
                flags, delta_v = struct.unpack(">bh", notification.data[:3])
                high_res = (flags & 0x10) != 0
                periods = flags & 0x0F
                logger.info("%s: WHEEL: res: %d periods: %d delta V:%-3d", device, high_res, periods, delta_v)
        elif notification.address == 0x10:
            ratchet = notification.data[0]
            if logger.isEnabledFor(logging.INFO):
                logger.info("%s: WHEEL: ratchet: %d", device, ratchet)
            if ratchet < 2:  # don't process messages with unusual ratchet values
                if device.setting_callback:
                    device.setting_callback(device, settings_templates.ScrollRatchet, [2 if ratchet else 1])
        else:
            if logger.isEnabledFor(logging.INFO):
                logger.info("%s: unknown WHEEL %s", device, notification)

    elif feature == SupportedFeature.ONBOARD_PROFILES:
        if notification.address > 0x10:
            if logger.isEnabledFor(logging.INFO):
                logger.info("%s: unknown ONBOARD PROFILES %s", device, notification)
        else:
            if notification.address == 0x00:
                profile_sector = struct.unpack("!H", notification.data[:2])[0]
                if profile_sector:
                    settings_templates.profile_change(device, profile_sector)
            elif notification.address == 0x10:
                resolution_index = struct.unpack("!B", notification.data[:1])[0]
                profile_sector = struct.unpack("!H", device.feature_request(SupportedFeature.ONBOARD_PROFILES, 0x40)[:2])[0]
                if device.setting_callback:
                    for profile in device.profiles.profiles.values() if device.profiles else []:
                        if profile.sector == profile_sector:
                            device.setting_callback(
                                device, settings_templates.AdjustableDpi, [profile.resolutions[resolution_index]]
                            )
                            break

    elif feature == SupportedFeature.BRIGHTNESS_CONTROL:
        if notification.address > 0x10:
            if logger.isEnabledFor(logging.INFO):
                logger.info("%s: unknown BRIGHTNESS CONTROL %s", device, notification)
        else:
            if notification.address == 0x00:
                brightness = struct.unpack("!H", notification.data[:2])[0]
                device.setting_callback(device, settings_templates.BrightnessControl, [brightness])
            elif notification.address == 0x10:
                brightness = notification.data[0] & 0x01
                if brightness:
                    brightness = struct.unpack("!H", device.feature_request(SupportedFeature.BRIGHTNESS_CONTROL, 0x10)[:2])[0]
                device.setting_callback(device, settings_templates.BrightnessControl, [brightness])

    diversion.process_notification(device, notification, feature)
    return True


def handle_pairing_lock(receiver: Receiver, notification: HIDPPNotification) -> bool:
    receiver.pairing.lock_open = bool(notification.address & 0x01)
    reason = _("pairing lock is open") if receiver.pairing.lock_open else _("pairing lock is closed")
    if logger.isEnabledFor(logging.INFO):
        logger.info("%s: %s", receiver, reason)
    receiver.pairing.error = None
    if receiver.pairing.lock_open:
        receiver.pairing.new_device = None
    pair_error = ord(notification.data[:1])
    if pair_error:
        error_string = hidpp10_constants.PairingError(pair_error).label
        receiver.pairing.error = error_string
        receiver.pairing.new_device = None
        logger.warning("pairing error %d: %s", pair_error, error_string)
    receiver.changed(reason=reason)
    return True


def handle_discovery_status(receiver: Receiver, notification: HIDPPNotification) -> bool:
    with notification_lock:
        receiver.pairing.discovering = notification.address == 0x00
        reason = _("discovery lock is open") if receiver.pairing.discovering else _("discovery lock is closed")
        if logger.isEnabledFor(logging.INFO):
            logger.info("%s: %s", receiver, reason)
        receiver.pairing.error = None
        if receiver.pairing.discovering:
            receiver.pairing.counter = receiver.pairing.device_address = None
            receiver.pairing.device_authentication = receiver.pairing.device_name = None
        receiver.pairing.device_passkey = None
        discover_error = ord(notification.data[:1])
        if discover_error:
            receiver.pairing.error = discover_string = hidpp10_constants.BoltPairingError(discover_error).label
            logger.warning("bolt discovering error %d: %s", discover_error, discover_string)
        receiver.changed(reason=reason)
        return True


def handle_device_discovery(receiver: Receiver, notification: HIDPPNotification) -> bool:
    with notification_lock:
        counter = notification.address + notification.data[0] * 256  # notification counter
        if receiver.pairing.counter is None:
            receiver.pairing.counter = counter
        else:
            if not receiver.pairing.counter == counter:
                return None
        if notification.data[1] == 0:
            receiver.pairing.device_kind = notification.data[3]
            receiver.pairing.device_address = notification.data[6:12]
            receiver.pairing.device_authentication = notification.data[14]
        elif notification.data[1] == 1:
            receiver.pairing.device_name = notification.data[3 : 3 + notification.data[2]].decode("utf-8")
        return True


def handle_pairing_status(receiver: Receiver, notification: HIDPPNotification) -> bool:
    with notification_lock:
        receiver.pairing.device_passkey = None
        receiver.pairing.lock_open = notification.address == 0x00
        reason = _("pairing lock is open") if receiver.pairing.lock_open else _("pairing lock is closed")
        if logger.isEnabledFor(logging.INFO):
            logger.info("%s: %s", receiver, reason)
        receiver.pairing.error = None
        if not receiver.pairing.lock_open:
            receiver.pairing.counter = None
            receiver.pairing.device_address = None
            receiver.pairing.device_authentication = None
            receiver.pairing.device_name = None
        pair_error = notification.data[0]
        if receiver.pairing.lock_open:
            receiver.pairing.new_device = None
        elif notification.address == 0x02 and not pair_error:
            receiver.pairing.new_device = receiver.register_new_device(notification.data[7])
        if pair_error:
            receiver.pairing.error = error_string = hidpp10_constants.BoltPairingError(pair_error).label
            receiver.pairing.new_device = None
            logger.warning("pairing error %d: %s", pair_error, error_string)
        receiver.changed(reason=reason)
        return True


def handle_passkey_request(receiver: Receiver, notification: HIDPPNotification) -> bool:
    with notification_lock:
        receiver.pairing.device_passkey = notification.data[0:6].decode("utf-8")
        return True


def handle_passkey_pressed(_receiver: Receiver, _hidpp_notification: HIDPPNotification) -> bool:
    return True



## Copyright (C) 2012-2013  Daniel Pavel
## Copyright (C) 2014-2024  Solaar Contributors https://pwr-solaar.github.io/Solaar/
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

import errno
import logging
import threading
import time
import typing

from typing import Callable
from typing import Optional
from typing import Protocol

from solaar import configuration

from . import descriptors
from . import exceptions
from . import hidpp10
from . import hidpp10_constants
from . import hidpp20
from . import settings
from . import settings_templates
from .common import Alert
from .common import Battery
from .hidpp10_constants import NotificationFlag
from .hidpp20_constants import SupportedFeature

if typing.TYPE_CHECKING:
    from logitech_receiver import common

logger = logging.getLogger(__name__)

_hidpp10 = hidpp10.Hidpp10()
_hidpp20 = hidpp20.Hidpp20()


class LowLevelInterface(Protocol):
    def open_path(self, path) -> int:
        ...

    def find_paired_node(self, receiver_path: str, index: int, timeout: int):
        ...

    def ping(self, handle, number, long_message: bool):
        ...

    def request(self, handle, devnumber, request_id, *params, **kwargs):
        ...

    def close(self, handle, *args, **kwargs) -> bool:
        ...


def create_device(low_level: LowLevelInterface, device_info, setting_callback=None):
    """Opens a Logitech Device found attached to the machine, by Linux device path.
    :returns: An open file handle for the found receiver, or None.
    """
    try:
        handle = low_level.open_path(device_info.path)
        if handle:
            # a direct connected device might not be online (as reported by user)
            return Device(
                low_level,
                None,
                None,
                None,
                handle=handle,
                device_info=device_info,
                setting_callback=setting_callback,
            )
    except OSError as e:
        logger.exception("open %s", device_info)
        if e.errno == errno.EACCES:
            raise e
    except Exception as e:
        logger.exception("open %s", device_info)
        raise e


class Device:
    instances = []
    read_register: Callable = hidpp10.read_register
    write_register: Callable = hidpp10.write_register

    def __init__(
        self,
        low_level: LowLevelInterface,
        receiver,
        number,
        online,
        pairing_info=None,
        handle=None,
        device_info=None,
        setting_callback=None,
    ):
        assert receiver or device_info
        if receiver:
            assert 0 < number <= 15  # some receivers have devices past their max # of devices
        self.low_level = low_level
        self.number = number  # will be None at this point for directly connected devices
        self.online = online  # is the device online? - gates many atempts to contact the device
        self.descriptor = None
        self.isDevice = True  # some devices act as receiver so we need a property to distinguish them
        self.may_unpair = False
        self.receiver = receiver
        self.handle = handle
        self.path = device_info.path if device_info else None
        self.product_id = device_info.product_id if device_info else None
        self.hidpp_short = device_info.hidpp_short if device_info else None
        self.hidpp_long = device_info.hidpp_long if device_info else None
        self.bluetooth = device_info.bus_id == 0x0005 if device_info else False  # Bluetooth needs long messages
        self.hid_serial = device_info.serial if device_info else None
        self.setting_callback = setting_callback  # for changes to settings
        self.status_callback = None  # for changes to other potentially visible aspects
        self.wpid = pairing_info["wpid"] if pairing_info else None  # the Wireless PID is unique per device model
        self._kind = pairing_info["kind"] if pairing_info else None  # mouse, keyboard, etc (see hidpp10.DEVICE_KIND)
        self._serial = pairing_info["serial"] if pairing_info else None  # serial number (an 8-char hex string)
        self._polling_rate = pairing_info["polling"] if pairing_info else None
        self._power_switch = pairing_info["power_switch"] if pairing_info else None
        self._name = None  # the full name of the model
        self._codename = None  # Unifying peripherals report a codename.
        self._protocol = None  # HID++ protocol version, 1.0 or 2.0
        self._unitId = None  # unit id (distinguishes within a model - generally the same as serial)
        self._modelId = None  # model id (contains identifiers for the transports of the device)
        self._tid_map = None  # map from transports to product identifiers
        self._persister = None  # persister holds settings
        self._led_effects = self._firmware = self._keys = self._remap_keys = self._gestures = self._force_buttons = None
        self._profiles = self._backlight = self._settings = None
        self.registers = []
        self.notification_flags = None
        self.battery_info = None
        self.link_encrypted = None
        self._active = None  # lags self.online - is used to help determine when to setup devices
        self.present = True  # used for devices that are integral with their receiver but that separately be disconnected

        self._feature_settings_checked = False
        self._gestures_lock = threading.Lock()
        self._settings_lock = threading.Lock()
        self._persister_lock = threading.Lock()
        self._notification_handlers = {}  # See `add_notification_handler`
        self.cleanups = []  # functions to run on the device when it is closed

        if not self.path:
            self.path = self.low_level.find_paired_node(receiver.path, number, 1) if receiver else None
        if not self.handle:
            try:
                self.handle = self.low_level.open_path(self.path) if self.path else None
            except Exception:  # maybe the device wasn't set up
                try:
                    time.sleep(1)
                    self.handle = self.low_level.open_path(self.path) if self.path else None
                except Exception:  # give up
                    self.handle = None  # should this give up completely?

        if receiver:
            if not self.wpid:
                raise exceptions.NoSuchDevice(
                    number=number, receiver=receiver, error="no wpid for device connected to receiver"
                )
            self.descriptor = descriptors.get_wpid(self.wpid)
            if self.descriptor is None:
                codename = self.receiver.device_codename(self.number)  # Last chance to get a descriptor, may fail
                if codename:
                    self._codename = codename
                    self.descriptor = descriptors.get_codename(self._codename)
        else:
            self.descriptor = (
                descriptors.get_btid(self.product_id) if self.bluetooth else descriptors.get_usbid(self.product_id)
            )
            if self.number is None:  # for direct-connected devices get 'number' from descriptor protocol else use 0xFF
                if self.descriptor and self.descriptor.protocol and self.descriptor.protocol < 2.0:
                    number = 0x00
                else:
                    number = 0xFF
                self.number = number
            self.ping()  # determine whether a direct-connected device is online

        if self.descriptor:
            self._name = self.descriptor.name
            if self._codename is None:
                self._codename = self.descriptor.codename
            if self._kind is None:
                self._kind = self.descriptor.kind
            self._protocol = self.descriptor.protocol if self.descriptor.protocol else None
            self.registers = self.descriptor.registers if self.descriptor.registers else []

        if self._protocol is not None:
            self.features = None if self._protocol < 2.0 else hidpp20.FeaturesArray(self)
        else:
            self.features = hidpp20.FeaturesArray(self)  # may be a 2.0 device; if not, it will fix itself later

        Device.instances.append(self)

    def find(self, id):  # find a device by serial number or unit ID or name or codename
        assert id, "need id to find a device"
        for device in Device.instances:
            if device.online and (device.unitId == id or device.serial == id or device.name == id or device.codename == id):
                return device

    @property
    def protocol(self):
        if not self._protocol:
            self.ping()
        return self._protocol or 0

    @property
    def codename(self):
        if not self._codename:
            if self.online and self.protocol >= 2.0:
                self._codename = _hidpp20.get_friendly_name(self)
                if not self._codename:
                    self._codename = self.name.split(" ", 1)[0] if self.name else None
            if not self._codename and self.receiver:
                codename = self.receiver.device_codename(self.number)
                if codename:
                    self._codename = codename
                elif self.protocol < 2.0:
                    self._codename = "? (%s)" % (self.wpid or self.product_id)
        return self._codename or f"?? ({self.wpid or self.product_id})"

    @property
    def name(self):
        if not self._name:
            if self.online and self.protocol >= 2.0:
                self._name = _hidpp20.get_name(self)
        return self._name or self._codename or f"Unknown device {self.wpid or self.product_id}"

    def get_ids(self):
        ids = _hidpp20.get_ids(self)
        if ids:
            self._unitId, self._modelId, self._tid_map = ids
            if logger.isEnabledFor(logging.INFO) and self._serial and self._serial != self._unitId:
                logger.info("%s: unitId %s does not match serial %s", self, self._unitId, self._serial)

    @property
    def unitId(self):
        if not self._unitId and self.online and self.protocol >= 2.0:
            self.get_ids()
        return self._unitId

    @property
    def modelId(self):
        if not self._modelId and self.online and self.protocol >= 2.0:
            self.get_ids()
        return self._modelId

    @property
    def tid_map(self):
        if not self._tid_map and self.online and self.protocol >= 2.0:
            self.get_ids()
        return self._tid_map

    @property
    def kind(self):
        if not self._kind and self.online and self.protocol >= 2.0:
            self._kind = _hidpp20.get_kind(self)
        return self._kind or "?"

    @property
    def firmware(self) -> tuple[common.FirmwareInfo]:
        if self._firmware is None and self.online:
            if self.protocol >= 2.0:
                self._firmware = _hidpp20.get_firmware(self)
            else:
                self._firmware = _hidpp10.get_firmware(self)
        return self._firmware or ()

    @property
    def serial(self):
        return self._serial or ""

    @property
    def id(self):
        return self.unitId or self.serial

    @property
    def power_switch_location(self):
        return self._power_switch

    @property
    def polling_rate(self):
        if self.online and self.protocol >= 2.0:
            rate = _hidpp20.get_polling_rate(self)
            self._polling_rate = rate if rate else self._polling_rate
        return self._polling_rate

    @property
    def led_effects(self):
        if not self._led_effects and self.online and self.protocol >= 2.0:
            if SupportedFeature.COLOR_LED_EFFECTS in self.features:
                self._led_effects = hidpp20.LEDEffectsInfo(self)
            elif SupportedFeature.RGB_EFFECTS in self.features:
                self._led_effects = hidpp20.RGBEffectsInfo(self)
        return self._led_effects

    @property
    def keys(self):
        if not self._keys:
            if self.online and self.protocol >= 2.0:
                self._keys = _hidpp20.get_keys(self) or ()
        return self._keys

    @property
    def remap_keys(self):
        if self._remap_keys is None:
            if self.online and self.protocol >= 2.0:
                self._remap_keys = _hidpp20.get_remap_keys(self) or ()
        return self._remap_keys

    @property
    def gestures(self):
        if self._gestures is None:
            with self._gestures_lock:
                if self._gestures is None:
                    if self.online and self.protocol >= 2.0:
                        self._gestures = _hidpp20.get_gestures(self) or ()
        return self._gestures

    @property
    def backlight(self):
        if self._backlight is None:
            if self.online and self.protocol >= 2.0:
                self._backlight = _hidpp20.get_backlight(self)
        return self._backlight

    @property
    def profiles(self):
        if self._profiles is None:
            if self.online and self.protocol >= 2.0:
                self._profiles = _hidpp20.get_profiles(self)
        return self._profiles

    def force_buttons(self):
        if self._force_buttons is None:
            if self.online and self.protocol >= 2.0:
                self._force_buttons = _hidpp20.get_force_buttons(self) or ()
        return self._force_buttons

    def set_configuration(self, configuration_, no_reply=False):
        if self.online and self.protocol >= 2.0:
            _hidpp20.config_change(self, configuration_, no_reply=no_reply)

    def reset(self, no_reply=False):
        self.set_configuration(0, no_reply)

    @property
    def persister(self):
        if not self._persister:
            with self._persister_lock:
                if not self._persister:
                    self._persister = configuration.persister(self)
        return self._persister

    @property
    def settings(self):
        if not self._settings:
            with self._settings_lock:
                if not self._settings:
                    settings = []
                    if self.persister and self.descriptor and self.descriptor.settings:
                        for sclass in self.descriptor.settings:
                            try:
                                setting = sclass.build(self)
                            except Exception as e:  # Do nothing if the device is offline
                                setting = None
                                if self.online:
                                    raise e
                            if setting is not None:
                                settings.append(setting)
                    self._settings = settings
        if not self._feature_settings_checked:
            with self._settings_lock:
                if not self._feature_settings_checked:
                    self._feature_settings_checked = settings_templates.check_feature_settings(self, self._settings)
        return self._settings

    def battery(self):  # None  or  level, next, status, voltage
        if self.protocol < 2.0:
            return _hidpp10.get_battery(self)
        else:
            battery_feature = self.persister.get("_battery", None) if self.persister else None
            if battery_feature != 0:
                result = _hidpp20.get_battery(self, battery_feature)
                try:
                    feature, battery = result
                    if self.persister and battery_feature is None:
                        self.persister["_battery"] = feature.value
                    return battery
                except Exception:
                    if self.persister and battery_feature is None and result is not None and result != 0:
                        self.persister["_battery"] = result.value

    def set_battery_info(self, info):
        """Update battery information for device, calling changed callback if necessary"""
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("%s: battery %s, %s", self, info.level, info.status)
        if info.level is None and self.battery_info:  # use previous level if missing from new information
            info.level = self.battery_info.level

        changed = self.battery_info != info
        self.battery_info, old_info = info, self.battery_info
        if old_info is None:
            old_info = Battery(None, None, None, None)

        alert, reason = Alert.NONE, None
        if not info.ok():
            logger.warning("%s: battery %d%%, ALERT %s", self, info.level, info.status)
            if old_info.status != info.status:
                alert = Alert.NOTIFICATION | Alert.ATTENTION
            reason = info.to_str()

        if changed or reason:
            # update the leds on the device, if any
            _hidpp10.set_3leds(self, info.level, charging=info.charging(), warning=bool(alert))
            self.changed(active=True, alert=alert, reason=reason)

    # Retrieve and regularize battery status
    def read_battery(self):
        if self.online:
            battery = self.battery()
            self.set_battery_info(battery if battery is not None else Battery(None, None, None, None))

    def changed(self, active=None, alert=Alert.NONE, reason=None, push=False):
        """The status of the device had changed, so invoke the status callback.
        Also push notifications and settings to the device when necessary."""
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("device %d changing: active=%s %s present=%s", self.number, active, self._active, self.present)
        if active is not None:
            self.online = active
            was_active, self._active = self._active, active
            if active:
                # Push settings for new devices when devices request software reconfiguration
                # and when devices become active if they don't have wireless device status feature,
                if (
                    was_active is None
                    or not was_active
                    or push
                    and (not self.features or SupportedFeature.WIRELESS_DEVICE_STATUS not in self.features)
                ):
                    if logger.isEnabledFor(logging.INFO):
                        logger.info("%s pushing device settings %s", self, self.settings)
                    settings.apply_all_settings(self)
                if not was_active:
                    if self.protocol < 2.0:  # Make sure to set notification flags on the device
                        self.notification_flags = self.enable_connection_notifications()
                    else:
                        self.set_configuration(0x11)  # signal end of configuration
                    self.read_battery()  # battery information may have changed so try to read it now
            elif was_active and self.receiver:  # need to set configuration pending flag in receiver
                hidpp10.set_configuration_pending_flags(self.receiver, 0xFF)
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("device %d changed: active=%s %s", self.number, self._active, self.battery_info)
        if self.status_callback is not None:
            self.status_callback(self, alert, reason)

    def enable_connection_notifications(self, enable=True):
        """Enable or disable device (dis)connection notifications on this
        receiver."""
        if not bool(self.receiver) or self.protocol >= 2.0:
            return False

        if enable:
            set_flag_bits = NotificationFlag.BATTERY_STATUS | NotificationFlag.UI | NotificationFlag.CONFIGURATION_COMPLETE
        else:
            set_flag_bits = 0
        ok = _hidpp10.set_notification_flags(self, set_flag_bits)
        if not ok:
            logger.warning("%s: failed to %s device notifications", self, "enable" if enable else "disable")

        flag_bits = _hidpp10.get_notification_flags(self)
        if logger.isEnabledFor(logging.INFO):
            if flag_bits is None:
                flag_names = None
            else:
                flag_names = hidpp10_constants.NotificationFlag.flag_names(flag_bits)
            is_enabled = "enabled" if enable else "disabled"
            logger.info(f"{self}: device notifications {is_enabled} {flag_names}")
        return flag_bits if ok else None

    def add_notification_handler(self, id: str, fn):
        """Adds the notification handling callback `fn` to this device under name `id`.
        If a callback has already been registered under this name, it's replaced with
        the argument.
        The callback will be invoked whenever the device emits an event message, and
        the resulting notification hasn't been handled by another handler on this device
        (order is not guaranteed, so handlers should not overlap in functionality).
        The callback should have type `(PairedDevice, Notification) -> Optional[bool]`.
        It should return `None` if it hasn't handled the notification, return `True`
        if it did so successfully and return `False` if an error should be reported
        (malformed notification, etc).
        """
        self._notification_handlers[id] = fn

    def remove_notification_handler(self, id: str):
        """Unregisters the notification handler under name `id`."""

        if id not in self._notification_handlers and logger.isEnabledFor(logging.INFO):
            logger.info(f"Tried to remove nonexistent notification handler {id} from device {self}.")
        else:
            del self._notification_handlers[id]

    def handle_notification(self, n) -> Optional[bool]:
        for h in self._notification_handlers.values():
            ret = h(self, n)
            if ret is not None:
                return ret
        return None

    def request(self, request_id, *params, no_reply=False):
        if self:
            long = self.hidpp_long is True or (
                self.hidpp_long is None and (self.bluetooth or self._protocol is not None and self._protocol >= 2.0)
            )
            return self.low_level.request(
                self.handle or (self.receiver.handle if self.receiver else None),
                self.number,
                request_id,
                *params,
                no_reply=no_reply,
                long_message=long,
                protocol=self.protocol,
            )

    def feature_request(self, feature, function=0x00, *params, no_reply=False):
        if self.protocol >= 2.0:
            return hidpp20.feature_request(self, feature, function, *params, no_reply=no_reply)

    def ping(self):
        """Checks if the device is online and present, returns True of False.
        Some devices are integral with their receiver but may not be present even if the receiver responds to ping."""
        long = self.hidpp_long is True or (
            self.hidpp_long is None and (self.bluetooth or self._protocol is not None and self._protocol >= 2.0)
        )
        handle = self.handle or self.receiver.handle
        try:
            protocol = self.low_level.ping(handle, self.number, long_message=long)
        except exceptions.NoReceiver:  # if ping fails, device is offline
            protocol = None
        self.online = protocol is not None and self.present
        if protocol:
            self._protocol = protocol
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("pinged %s: online %s protocol %s present %s", self.number, self.online, protocol, self.present)
        return self.online

    def notify_devices(self):  # no need to notify, as there are none
        pass

    def close(self):
        handle, self.handle = self.handle, None
        if self in Device.instances:
            Device.instances.remove(self)
        if hasattr(self, "cleanups"):
            for cleanup in self.cleanups:
                cleanup(self)
        return handle and self.low_level.close(handle)

    def __index__(self):
        return self.number

    __int__ = __index__

    def __eq__(self, other):
        return other is not None and self._kind == other._kind and self.wpid == other.wpid

    def __ne__(self, other):
        return other is None or self.kind != other.kind or self.wpid != other.wpid

    def __hash__(self):
        return self.wpid.__hash__()

    def __bool__(self):
        return self.wpid is not None and self.number in self.receiver if self.receiver else self.handle is not None

    __nonzero__ = __bool__

    def status_string(self):
        return self.battery_info.to_str() if self.battery_info is not None else ""

    def __str__(self):
        try:
            name = self._name or self._codename or "?"
        except exceptions.NoSuchDevice:
            name = "name not available"
        return f"<Device({int(self.number)},{self.wpid or self.product_id},{name},{self.serial})>"

    __repr__ = __str__

    def __del__(self):
        self.close()



## Copyright (C) 2012-2013  Daniel Pavel
## Copyright (C) 2014-2024  Solaar Contributors https://pwr-solaar.github.io/Solaar/
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import logging
import os

from enum import Enum
from time import time

import gi

from gi.repository import GLib
from gi.repository import Gtk
from gi.repository.Gdk import ScrollDirection

import solaar.gtk as gtk

from solaar import NAME
from solaar.i18n import _

from . import action
from . import icons
from . import window
from .about import about

logger = logging.getLogger(__name__)

_TRAY_ICON_SIZE = 48
_MENU_ICON_SIZE = Gtk.IconSize.LARGE_TOOLBAR


class GtkSignal(Enum):
    ACTIVATE = "activate"
    SCROLL_EVENT = "scroll-event"


def _create_menu(quit_handler):
    # per-device menu entries will be generated as-needed
    menu = Gtk.Menu()

    no_receiver = Gtk.MenuItem.new_with_label(_("No supported device found"))
    no_receiver.set_sensitive(False)
    menu.append(no_receiver)
    menu.append(Gtk.SeparatorMenuItem.new())

    menu.append(action.make_image_menu_item(_("About %s") % NAME, "help-about", about.show))
    menu.append(action.make_image_menu_item(_("Quit %s") % NAME, "application-exit", quit_handler))

    menu.show_all()
    return menu


_last_scroll = 0


def _scroll(tray_icon, event, direction=None):
    if direction is None:
        direction = event.direction
        now = event.time / 1000.0
    else:
        now = None

    if direction != ScrollDirection.UP and direction != ScrollDirection.DOWN:
        # ignore all other directions
        return

    # don't bother even trying to scroll if less than two devices
    if sum(map(lambda i: i[1] is not None, _devices_info)) < 2:
        return

    # scroll events come way too fast (at least 5-6 at once) so take a little break between them
    global _last_scroll
    now = now or time()
    if now - _last_scroll < 0.33:  # seconds
        return
    _last_scroll = now

    global _picked_device
    candidate = None

    if _picked_device is None:
        for info in _devices_info:
            # pick first peripheral found
            if info[1] is not None:
                candidate = info
                break
    else:
        found = False
        for info in _devices_info:
            if not info[1]:
                # only conside peripherals
                continue
            # compare peripherals
            if info[0:2] == _picked_device[0:2]:
                if direction == ScrollDirection.UP and candidate:
                    # select previous device
                    break
                found = True
            else:
                if found:
                    candidate = info
                    if direction == ScrollDirection.DOWN:
                        break
                    # if direction is up, but no candidate found before _picked,
                    # let it run through all candidates, will get stuck with the last one
                else:
                    if direction == ScrollDirection.DOWN:
                        # only use the first one, in case no candidates are after _picked
                        if candidate is None:
                            candidate = info
                    else:
                        candidate = info

        # if the last _picked_device is gone, clear it
        # the candidate will be either the first or last one remaining,
        # depending on the scroll direction
        if not found:
            _picked_device = None

    _picked_device = candidate or _picked_device
    logger.debug("scroll: picked %s", _picked_device)
    _update_tray_icon()


try:
    try:
        gi.require_version("AyatanaAppIndicator3", "0.1")
        from gi.repository import AyatanaAppIndicator3 as AppIndicator3

        ayatana_appindicator_found = True
    except ValueError:
        try:
            gi.require_version("AppIndicator3", "0.1")
            from gi.repository import AppIndicator3

            ayatana_appindicator_found = False
        except ValueError as exc:
            # treat unavailable versions the same as unavailable packages
            raise ImportError from exc

    logger.debug(f"using {'Ayatana ' if ayatana_appindicator_found else ''}AppIndicator3")

    # Defense against AppIndicator3 bug that treats files in current directory as icon files
    # https://bugs.launchpad.net/ubuntu/+source/libappindicator/+bug/1363277
    # Defense against bug that shows up in XFCE 4.16 where icons are not upscaled
    def _icon_file(icon_name):
        if gtk.tray_icon_size is None and not os.path.isfile(icon_name):
            return icon_name
        icon_info = Gtk.IconTheme.get_default().lookup_icon(
            icon_name, gtk.tray_icon_size or _TRAY_ICON_SIZE, Gtk.IconLookupFlags.FORCE_SVG
        )
        return icon_info.get_filename() if icon_info else icon_name

    def _create(menu):
        icons._init_icon_paths()
        ind = AppIndicator3.Indicator.new(
            "indicator-solaar", _icon_file(icons.TRAY_INIT), AppIndicator3.IndicatorCategory.HARDWARE
        )
        ind.set_title(NAME)
        ind.set_status(AppIndicator3.IndicatorStatus.ACTIVE)
        # ind.set_attention_icon_full(_icon_file(icons.TRAY_ATTENTION), '') # works poorly for XFCE 16
        # ind.set_label(NAME.lower(), NAME.lower())

        ind.set_menu(menu)
        ind.connect(GtkSignal.SCROLL_EVENT.value, _scroll)

        return ind

    def _hide(indicator):
        indicator.set_status(AppIndicator3.IndicatorStatus.PASSIVE)

    def _show(indicator):
        indicator.set_status(AppIndicator3.IndicatorStatus.ACTIVE)

    def _update_tray_icon():
        if _picked_device and gtk.battery_icons_style != "solaar":
            _ignore, _ignore, name, device = _picked_device
            battery_level = device.battery_info.level if device.battery_info is not None else None
            battery_charging = device.battery_info.charging() if device.battery_info is not None else None
            tray_icon_name = icons.battery(battery_level, battery_charging)
            description = f"{name}: {device.status_string()}"
        else:
            # there may be a receiver, but no peripherals
            tray_icon_name = icons.TRAY_OKAY if _devices_info else icons.TRAY_INIT

            description_lines = _generate_description_lines()
            description = "\n".join(description_lines).rstrip("\n")

        # icon_file = icons.icon_file(icon_name, _TRAY_ICON_SIZE)
        _icon.set_icon_full(_icon_file(tray_icon_name), description)

    def attention(reason=None):
        if _icon.get_status() != AppIndicator3.IndicatorStatus.ATTENTION:
            # _icon.set_attention_icon_full(_icon_file(icons.TRAY_ATTENTION), reason or '') # works poorly for XFCe 16
            _icon.set_status(AppIndicator3.IndicatorStatus.ATTENTION)
            GLib.timeout_add(10 * 1000, _icon.set_status, AppIndicator3.IndicatorStatus.ACTIVE)

except ImportError:
    logger.debug("using StatusIcon")

    def _create(menu):
        icon = Gtk.StatusIcon.new_from_icon_name(icons.TRAY_INIT)
        icon.set_name(NAME.lower())
        icon.set_title(NAME)
        icon.set_tooltip_text(NAME)
        icon.connect(GtkSignal.ACTIVATE.value, window.toggle)
        icon.connect(GtkSignal.SCROLL_EVENT.value, _scroll)
        icon.connect(
            "popup-menu",
            lambda icon, button, time: menu.popup(None, None, icon.position_menu, icon, button, time),
        )

        return icon

    def _hide(icon):
        icon.set_visible(False)

    def _show(icon):
        icon.set_visible(True)

    def _update_tray_icon():
        tooltip_lines = _generate_tooltip_lines()
        tooltip = "\n".join(tooltip_lines).rstrip("\n")
        _icon.set_tooltip_markup(tooltip)

        if _picked_device and gtk.battery_icons_style != "solaar":
            _ignore, _ignore, name, device = _picked_device
            battery_level = device.battery_info.level if device.battery_info is not None else None
            battery_charging = device.battery_info.charging() if device.battery_info is not None else None
            tray_icon_name = icons.battery(battery_level, battery_charging)
        else:
            # there may be a receiver, but no peripherals
            tray_icon_name = icons.TRAY_OKAY if _devices_info else icons.TRAY_ATTENTION
        _icon.set_from_icon_name(tray_icon_name)

    _icon_before_attention = None

    def _blink(count):
        global _icon_before_attention
        if count % 2:
            _icon.set_from_icon_name(icons.TRAY_ATTENTION)
        else:
            _icon.set_from_icon_name(_icon_before_attention)

        if count > 0:
            GLib.timeout_add(1000, _blink, count - 1)
        else:
            _icon_before_attention = None

    def attention(reason=None):
        global _icon_before_attention
        if _icon_before_attention is None:
            _icon_before_attention = _icon.get_icon_name()
            GLib.idle_add(_blink, 9)


def _generate_tooltip_lines():
    if not _devices_info:
        yield f"<b>{NAME}</b>: " + _("no receiver")
        return

    yield from _generate_description_lines()


def _generate_description_lines():
    if not _devices_info:
        yield _("no receiver")
        return

    for _ignore, number, name, device in _devices_info:
        if number is None:  # receiver
            continue

        p = device.status_string()
        if p:  # does it have any properties to print?
            yield f"<b>{name}</b>"
            if device.online:
                yield f"\t{p}"
            else:
                yield f"\t{p} <small>(" + _("offline") + ")</small>"
        else:
            if device.online:
                yield f"<b>{name}</b> <small>(" + _("no status") + ")</small>"
            else:
                yield f"<b>{name}</b> <small>(" + _("offline") + ")</small>"


def _pick_device_with_lowest_battery():
    if not _devices_info:
        return None

    picked = None
    picked_level = 1000

    for info in _devices_info:
        if info[1] is None:  # is receiver
            continue
        level = info[-1].battery_info.level if info[-1].battery_info is not None else None
        if level is not None and picked_level > level:
            picked = info
            picked_level = level or 0

    logger.debug("picked device with lowest battery: %s", picked)

    return picked


def _add_device(device):
    assert device

    index = 0
    receiver_path = device.receiver.path if device.receiver is not None else device.path
    if device.receiver is not None:  # if receiver insert into devices for the receiver in device number order
        for idx, (path, _ignore, _ignore, _ignore) in enumerate(_devices_info):
            if path and path == receiver_path:
                index = idx + 1  # the first entry matching the receiver serial should be for the receiver itself
                break
        while index < len(_devices_info):
            path, number, _ignore, _ignore = _devices_info[index]
            if not path == receiver_path:
                break
            assert number != device.number
            if number > device.number:
                break
            index = index + 1

    new_device_info = (receiver_path, device.number, device.name, device)
    _devices_info.insert(index, new_device_info)

    label = ("   " if device.number else "") + device.name
    new_menu_item = action.make_image_menu_item(label, None, window.popup, receiver_path, device.number)
    _menu.insert(new_menu_item, index)

    return index


def _remove_device(index):
    assert index is not None

    menu_items = _menu.get_children()
    _menu.remove(menu_items[index])

    removed_device = _devices_info.pop(index)
    global _picked_device
    if _picked_device and _picked_device[0:2] == removed_device[0:2]:
        # the current pick was unpaired
        _picked_device = None


def _add_receiver(receiver):
    index = len(_devices_info)
    new_receiver_info = (receiver.path, None, receiver.name, None)
    _devices_info.insert(index, new_receiver_info)
    icon_name = icons.device_icon_name(receiver.name, receiver.kind)
    new_menu_item = action.make_image_menu_item(receiver.name, icon_name, window.popup, receiver.path)
    _menu.insert(new_menu_item, index)
    return 0


def _remove_receiver(receiver):
    index = 0
    # remove all entries in devices_info that match this receiver
    while index < len(_devices_info):
        path, _ignore, _ignore, _ignore = _devices_info[index]
        if path == receiver.path:
            _remove_device(index)
        else:
            index += 1


def _update_menu_item(index, device):
    if device is None:
        logger.warning("updating an inactive device %s, assuming disconnected", device)
        return None
    menu_items = _menu.get_children()
    menu_item = menu_items[index]
    level = device.battery_info.level if device.battery_info is not None else None
    charging = device.battery_info.charging() if device.battery_info is not None else None
    icon_name = icons.battery(level, charging)
    menu_item.label.set_label(("  " if 0 < device.number <= 6 else "") + device.name + ": " + device.status_string())
    image_widget = menu_item.icon
    image_widget.set_sensitive(bool(device.online))
    image_widget.set_from_icon_name(icon_name, _MENU_ICON_SIZE)


# for which device to show the battery info in systray, if more than one
# it's actually an entry in _devices_info
_picked_device = None

# cached list of devices and some of their properties
# contains tuples of (receiver path, device number, name, device)
_devices_info = []

_menu = None
_icon = None


def init(_quit_handler):
    global _menu, _icon
    assert _menu is None
    _menu = _create_menu(_quit_handler)
    assert _icon is None
    _icon = _create(_menu)
    update()


def destroy():
    global _icon, _menu, _devices_info
    if _icon is not None:
        i, _icon = _icon, None
        _hide(i)
        i = None

    _icon = None
    _menu = None
    _devices_info = None


def update(device=None):
    if _icon is None:
        return

    if device is not None:
        if device.kind is None:
            # receiver
            is_alive = bool(device)
            receiver_path = device.path
            if is_alive:
                index = None
                for idx, (path, _ignore, _ignore, _ignore) in enumerate(_devices_info):
                    if path == receiver_path:
                        index = idx
                        break

                if index is None:
                    _add_receiver(device)
            else:
                _remove_receiver(device)

        else:
            # peripheral
            is_paired = bool(device)
            receiver_path = device.receiver.path if device.receiver is not None else device.path
            index = None
            for idx, (path, number, _ignore, _ignore) in enumerate(_devices_info):
                if path == receiver_path and number == device.number:
                    index = idx

            if is_paired:
                if index is None:
                    index = _add_device(device)
                _update_menu_item(index, device)
            else:  # was just unpaired or unplugged
                if index is not None:
                    _remove_device(index)

        menu_items = _menu.get_children()
        no_receivers_index = len(_devices_info)
        menu_items[no_receivers_index].set_visible(not _devices_info)

    global _picked_device
    if (not _picked_device or _last_scroll == 0) and device is not None and device.kind is not None:
        # if it's just a receiver update, it's unlikely the picked device would change
        _picked_device = _pick_device_with_lowest_battery()

    _update_tray_icon()

    if _icon:
        if not _devices_info:
            _hide(_icon)
        else:
            _show(_icon)



## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import pkgutil
import subprocess
import sys

NAME = "Solaar"

try:
    __version__ = (
        subprocess.check_output(
            [
                "git",
                "describe",
                "--always",
            ],
            cwd=sys.path[0],
            stderr=subprocess.DEVNULL,
        )
        .strip()
        .decode()
    )
except Exception:
    try:
        __version__ = pkgutil.get_data("solaar", "commit").strip().decode()
    except Exception:
        __version__ = pkgutil.get_data("solaar", "version").strip().decode()



## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import logging

from enum import Enum
from typing import Tuple

import gi

from solaar.i18n import _
from solaar.tasks import TaskRunner

gi.require_version("Gtk", "3.0")
from gi.repository import GLib  # NOQA: E402
from gi.repository import Gtk  # NOQA: E402

logger = logging.getLogger(__name__)


class ErrorReason(Enum):
    PERMISSIONS = "Permissions"
    NO_DEVICE = "No device"
    UNPAIR = "Unpair"


def _create_error_text(reason: ErrorReason, object_) -> Tuple[str, str]:
    if reason == ErrorReason.PERMISSIONS:
        title = _("Permissions error")
        text = (
            _("Found a Logitech receiver or device (%s), but did not have permission to open it.") % object_
            + "\n\n"
            + _("If you've just installed Solaar, try disconnecting the receiver or device and then reconnecting it.")
        )
    elif reason == ErrorReason.NO_DEVICE:
        title = _("Cannot connect to device error")
        text = (
            _("Found a Logitech receiver or device at %s, but encountered an error connecting to it.") % object_
            + "\n\n"
            + _("Try disconnecting the device and then reconnecting it or turning it off and then on.")
        )
    elif reason == ErrorReason.UNPAIR:
        title = _("Unpairing failed")
        text = (
            _("Failed to unpair %{device} from %{receiver}.").format(
                device=object_.name,
                receiver=object_.receiver.name,
            )
            + "\n\n"
            + _("The receiver returned an error, with no further details.")
        )
    else:
        raise Exception("ui.error_dialog: don't know how to handle (%s, %s)", reason.name, object_)
    return title, text


def _error_dialog(reason: ErrorReason, object_):
    logger.error("error: %s %s", reason, object_)
    title, text = _create_error_text(reason, object_)

    m = Gtk.MessageDialog(None, Gtk.DialogFlags.MODAL, Gtk.MessageType.ERROR, Gtk.ButtonsType.CLOSE, text)
    m.set_title(title)
    m.run()
    m.destroy()


def error_dialog(reason: ErrorReason, object_):
    GLib.idle_add(_error_dialog, reason, object_)


_task_runner = None


def start_async():
    global _task_runner
    _task_runner = TaskRunner("AsyncUI")
    _task_runner.start()


def stop_async():
    global _task_runner
    _task_runner.stop()
    _task_runner = None


def ui_async(function, *args, **kwargs):
    """Runs a function asynchronously."""
    if _task_runner:
        _task_runner(function, *args, **kwargs)



from __future__ import annotations

import dataclasses


@dataclasses.dataclass
class DeviceInfo:
    path: str
    bus_id: str | None
    vendor_id: str
    product_id: str
    interface: str | None
    driver: str | None
    manufacturer: str | None
    product: str | None
    serial: str | None
    release: str | None
    isDevice: bool
    hidpp_short: str | None
    hidpp_long: str | None



## Copyright (C) 2012-2013  Daniel Pavel
## Copyright (C) 2014-2024  Solaar Contributors https://pwr-solaar.github.io/Solaar/
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import logging

from enum import Enum
from enum import IntEnum

import gi

from gi.repository.GObject import TYPE_PYOBJECT
from logitech_receiver import hidpp10_constants
from logitech_receiver.common import LOGITECH_VENDOR_ID
from logitech_receiver.common import NamedInt

from solaar import NAME
from solaar.i18n import _
from solaar.i18n import ngettext

from . import action
from . import config_panel
from . import diversion_rules
from . import icons
from .about import about
from .common import ui_async

gi.require_version("Gdk", "3.0")
from gi.repository import Gdk  # NOQA: E402
from gi.repository import GLib  # NOQA: E402
from gi.repository import Gtk  # NOQA: E402

logger = logging.getLogger(__name__)

_SMALL_BUTTON_ICON_SIZE = Gtk.IconSize.MENU
_NORMAL_BUTTON_ICON_SIZE = Gtk.IconSize.BUTTON
_TREE_ICON_SIZE = Gtk.IconSize.BUTTON
_INFO_ICON_SIZE = Gtk.IconSize.LARGE_TOOLBAR
_DEVICE_ICON_SIZE = Gtk.IconSize.DND
try:
    gi.check_version("3.7.4")
    _CAN_SET_ROW_NONE = None
except (ValueError, AttributeError):
    _CAN_SET_ROW_NONE = ""


class Column(IntEnum):
    """Columns of tree model."""

    PATH = 0
    NUMBER = 1
    ACTIVE = 2
    NAME = 3
    ICON = 4
    STATUS_TEXT = 5
    STATUS_ICON = 6
    DEVICE = 7


_COLUMN_TYPES = (str, int, bool, str, str, str, str, TYPE_PYOBJECT)
_TREE_SEPATATOR = (None, 0, False, None, None, None, None, None)
assert len(_TREE_SEPATATOR) == len(_COLUMN_TYPES)
assert len(_COLUMN_TYPES) == len(Column)


class GtkSignal(Enum):
    CHANGED = "changed"
    CLICKED = "clicked"
    DELETE_EVENT = "delete-event"


def _new_button(label, icon_name=None, icon_size=_NORMAL_BUTTON_ICON_SIZE, tooltip=None, toggle=False, clicked=None):
    b = Gtk.ToggleButton() if toggle else Gtk.Button()
    c = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 5)
    if icon_name:
        c.pack_start(Gtk.Image.new_from_icon_name(icon_name, icon_size), True, True, 0)
    if label:
        c.pack_start(Gtk.Label(label=label), True, True, 0)
    b.add(c)
    if clicked is not None:
        b.connect(GtkSignal.CLICKED.value, clicked)
    if tooltip:
        b.set_tooltip_text(tooltip)
    if not label and icon_size < _NORMAL_BUTTON_ICON_SIZE:
        b.set_relief(Gtk.ReliefStyle.NONE)
        b.set_focus_on_click(False)
    return b


def _create_receiver_panel():
    p = Gtk.Box.new(Gtk.Orientation.VERTICAL, 4)

    p._count = Gtk.Label()
    p._count.set_margin_top(24)
    p.pack_start(p._count, True, True, 0)

    p._scanning = Gtk.Label(label=_("Scanning") + "...")
    p._spinner = Gtk.Spinner()

    bp = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 8)
    bp.pack_start(Gtk.Label(label=" "), True, True, 0)
    bp.pack_start(p._scanning, False, False, 0)
    bp.pack_end(p._spinner, False, False, 0)
    p.pack_end(bp, False, False, 0)

    return p


def _create_device_panel():
    p = Gtk.Box.new(Gtk.Orientation.VERTICAL, 4)

    def _status_line(label_text):
        b = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 8)
        b.set_size_request(10, 28)

        b._label = Gtk.Label(label=label_text)
        b._label.set_size_request(170, 10)
        b.pack_start(b._label, False, False, 0)

        b._icon = Gtk.Image()
        b.pack_start(b._icon, False, False, 0)

        b._text = Gtk.Label()
        b.pack_start(b._text, False, False, 0)

        return b

    p._battery = _status_line(_("Battery"))
    p.pack_start(p._battery, False, False, 0)

    p._secure = _status_line(_("Wireless Link"))
    p._secure._icon.set_from_icon_name("dialog-warning", _INFO_ICON_SIZE)
    p.pack_start(p._secure, False, False, 0)

    p._lux = _status_line(_("Lighting"))
    p.pack_start(p._lux, False, False, 0)

    p.pack_start(Gtk.Separator.new(Gtk.Orientation.HORIZONTAL), False, False, 0)  # spacer

    p._config = config_panel.create()
    p.pack_end(p._config, True, True, 4)

    return p


def _create_details_panel():
    p = Gtk.Frame()
    p.set_shadow_type(Gtk.ShadowType.NONE)
    p.set_size_request(240, 0)
    p.set_state_flags(Gtk.StateFlags.ACTIVE, True)

    p._text = Gtk.Label()
    p._text.set_margin_start(6)
    p._text.set_margin_end(4)
    p._text.set_selectable(True)
    p.add(p._text)

    return p


def _create_buttons_box():
    bb = Gtk.HButtonBox()
    bb.set_layout(Gtk.ButtonBoxStyle.END)

    bb._details = _new_button(
        None,
        "dialog-information",
        _SMALL_BUTTON_ICON_SIZE,
        tooltip=_("Show Technical Details"),
        toggle=True,
        clicked=_update_details,
    )
    bb.add(bb._details)
    bb.set_child_secondary(bb._details, True)
    bb.set_child_non_homogeneous(bb._details, True)

    def _pair_new_device(trigger):
        assert _find_selected_device_id() is not None
        receiver = _find_selected_device()
        assert receiver is not None
        assert bool(receiver)
        assert receiver.kind is None
        action.pair(_window, receiver)

    bb._pair = _new_button(_("Pair new device"), "list-add", clicked=_pair_new_device)
    bb.add(bb._pair)

    def _unpair_current_device(trigger):
        assert _find_selected_device_id() is not None
        device = _find_selected_device()
        assert device is not None
        assert device.kind is not None
        action.unpair(_window, device)

    bb._unpair = _new_button(_("Unpair"), "edit-delete", clicked=_unpair_current_device)
    bb.add(bb._unpair)

    return bb


def _create_empty_panel():
    p = Gtk.Label()
    p.set_markup("<small>" + _("Select a device") + "</small>")
    p.set_sensitive(False)

    return p


def _create_info_panel():
    p = Gtk.Box.new(Gtk.Orientation.VERTICAL, 4)

    p._title = Gtk.Label(label=" ")
    p._icon = Gtk.Image()

    b1 = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 4)
    b1.pack_start(p._title, True, True, 0)
    b1.pack_start(p._icon, False, False, 0)
    p.pack_start(b1, False, False, 0)

    p.pack_start(Gtk.Separator.new(Gtk.Orientation.HORIZONTAL), False, False, 0)  # spacer

    p._receiver = _create_receiver_panel()
    p.pack_start(p._receiver, True, True, 0)

    p._device = _create_device_panel()
    p.pack_start(p._device, True, True, 0)

    p.pack_start(Gtk.Separator.new(Gtk.Orientation.HORIZONTAL), False, False, 0)  # spacer

    p._buttons = _create_buttons_box()
    p.pack_end(p._buttons, False, False, 0)

    return p


def _create_tree(model):
    tree = Gtk.TreeView()
    tree.set_size_request(330, 0)  # enough width for simple setups
    tree.set_headers_visible(False)
    tree.set_show_expanders(False)
    tree.set_level_indentation(20)
    tree.set_enable_tree_lines(True)
    tree.set_reorderable(False)
    tree.set_enable_search(False)
    tree.set_model(model)

    def _is_separator(model, item, _ignore=None):
        return model.get_value(item, Column.PATH) is None

    tree.set_row_separator_func(_is_separator, None)

    icon_cell_renderer = Gtk.CellRendererPixbuf()
    icon_cell_renderer.set_property("stock-size", _TREE_ICON_SIZE)
    icon_column = Gtk.TreeViewColumn("Icon", icon_cell_renderer)
    icon_column.add_attribute(icon_cell_renderer, "sensitive", Column.ACTIVE)
    icon_column.add_attribute(icon_cell_renderer, "icon-name", Column.ICON)
    tree.append_column(icon_column)

    name_cell_renderer = Gtk.CellRendererText()
    name_column = Gtk.TreeViewColumn("device name", name_cell_renderer)
    name_column.add_attribute(name_cell_renderer, "sensitive", Column.ACTIVE)
    name_column.add_attribute(name_cell_renderer, "text", Column.NAME)
    name_column.set_expand(True)
    tree.append_column(name_column)
    tree.set_expander_column(name_column)

    status_cell_renderer = Gtk.CellRendererText()
    status_cell_renderer.set_property("scale", 0.85)
    status_cell_renderer.set_property("xalign", 1)
    status_column = Gtk.TreeViewColumn("status text", status_cell_renderer)
    status_column.add_attribute(status_cell_renderer, "sensitive", Column.ACTIVE)
    status_column.add_attribute(status_cell_renderer, "text", Column.STATUS_TEXT)
    status_column.set_expand(True)
    tree.append_column(status_column)

    battery_cell_renderer = Gtk.CellRendererPixbuf()
    battery_cell_renderer.set_property("stock-size", _TREE_ICON_SIZE)
    battery_column = Gtk.TreeViewColumn("status icon", battery_cell_renderer)
    battery_column.add_attribute(battery_cell_renderer, "sensitive", Column.ACTIVE)
    battery_column.add_attribute(battery_cell_renderer, "icon-name", Column.STATUS_ICON)
    tree.append_column(battery_column)

    return tree


def _create_window_layout():
    assert _tree is not None
    assert _details is not None
    assert _info is not None
    assert _empty is not None

    assert _tree.get_selection().get_mode() == Gtk.SelectionMode.SINGLE
    _tree.get_selection().connect(GtkSignal.CHANGED.value, _device_selected)

    tree_scroll = Gtk.ScrolledWindow()
    tree_scroll.add(_tree)
    tree_scroll.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
    tree_scroll.set_shadow_type(Gtk.ShadowType.IN)

    tree_panel = Gtk.Box.new(Gtk.Orientation.VERTICAL, 0)
    tree_panel.set_homogeneous(False)
    tree_panel.pack_start(tree_scroll, True, True, 0)
    tree_panel.pack_start(_details, False, False, 0)

    panel = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 16)
    panel.pack_start(tree_panel, False, False, 0)
    panel.pack_start(_info, True, True, 0)
    panel.pack_start(_empty, True, True, 0)

    bottom_buttons_box = Gtk.HButtonBox()
    bottom_buttons_box.set_layout(Gtk.ButtonBoxStyle.START)
    bottom_buttons_box.set_spacing(20)
    quit_button = _new_button(_("Quit %s") % NAME, "application-exit", _SMALL_BUTTON_ICON_SIZE, clicked=destroy)
    bottom_buttons_box.add(quit_button)
    about_button = _new_button(_("About %s") % NAME, "help-about", _SMALL_BUTTON_ICON_SIZE, clicked=about.show)
    bottom_buttons_box.add(about_button)
    diversion_button = _new_button(
        _("Rule Editor"), "", _SMALL_BUTTON_ICON_SIZE, clicked=lambda *_trigger: diversion_rules.show_window(_model)
    )
    bottom_buttons_box.add(diversion_button)
    bottom_buttons_box.set_child_secondary(diversion_button, True)

    vbox = Gtk.Box.new(Gtk.Orientation.VERTICAL, 8)
    vbox.set_border_width(8)
    vbox.pack_start(panel, True, True, 0)
    vbox.pack_end(bottom_buttons_box, False, False, 0)
    vbox.show_all()

    _details.set_visible(False)
    _info.set_visible(False)
    return vbox


def _create(delete_action):
    window = Gtk.Window()
    window.set_title(NAME)
    window.set_role("status-window")
    window.connect(GtkSignal.DELETE_EVENT.value, delete_action)

    vbox = _create_window_layout()
    window.add(vbox)

    geometry = Gdk.Geometry()
    geometry.min_width = 600
    geometry.min_height = 320
    window.set_geometry_hints(vbox, geometry, Gdk.WindowHints.MIN_SIZE)
    window.set_position(Gtk.WindowPosition.CENTER)

    style = window.get_style_context()
    style.add_class("solaar")

    return window


def _find_selected_device():
    selection = _tree.get_selection()
    model, item = selection.get_selected()
    return model.get_value(item, Column.DEVICE) if item else None


def _find_selected_device_id():
    selection = _tree.get_selection()
    model, item = selection.get_selected()
    if item:
        return _model.get_value(item, Column.PATH), _model.get_value(item, Column.NUMBER)


# triggered by changing selection in the tree
def _device_selected(selection):
    model, item = selection.get_selected()
    device = model.get_value(item, Column.DEVICE) if item else None
    if device:
        _update_info_panel(device, full=True)
    else:
        # When removing a receiver, one of its children may get automatically selected
        # before the tree had time to remove them as well.
        # Rather than chase around for another device to select, just clear the selection.
        _tree.get_selection().unselect_all()
        _update_info_panel(None, full=True)


def _receiver_row(receiver_path, receiver=None):
    assert receiver_path
    r = _model.get_iter_first()
    while r:
        r = _model.iter_next(r)

    item = _model.get_iter_first()
    while item:
        # first row matching the path must be the receiver one
        if _model.get_value(item, Column.PATH) == receiver_path:
            return item
        item = _model.iter_next(item)

    if not item and receiver:
        icon_name = icons.device_icon_name(receiver.name)
        status_text = None
        status_icon = None
        row_data = (receiver_path, 0, True, receiver.name, icon_name, status_text, status_icon, receiver)
        assert len(row_data) == len(_TREE_SEPATATOR)
        logger.debug("new receiver row %s", row_data)
        item = _model.append(None, row_data)
        if _TREE_SEPATATOR:
            _model.append(None, _TREE_SEPATATOR)

    return item or None


def _device_row(receiver_path, device_number, device=None):
    assert receiver_path
    assert device_number is not None

    receiver_row = _receiver_row(receiver_path, None if device is None else device.receiver)

    if device_number == 0xFF or device_number == 0x0:  # direct-connected device, receiver row is device row
        if receiver_row:
            return receiver_row
        item = None
        new_child_index = 0
    else:
        item = _model.iter_children(receiver_row)
        new_child_index = 0
        while item:
            if _model.get_value(item, Column.PATH) != receiver_path:
                logger.warning(
                    "path for device row %s different from path for receiver %s",
                    _model.get_value(item, Column.PATH),
                    receiver_path,
                )
            item_number = _model.get_value(item, Column.NUMBER)
            if item_number == device_number:
                return item
            if item_number > device_number:
                item = None
                break
            new_child_index += 1
            item = _model.iter_next(item)

    if not item and device:
        icon_name = icons.device_icon_name(device.name, device.kind)
        status_text = None
        status_icon = None
        row_data = (
            receiver_path,
            device_number,
            bool(device.online),
            device.codename,
            icon_name,
            status_text,
            status_icon,
            device,
        )
        assert len(row_data) == len(_TREE_SEPATATOR)
        logger.debug("new device row %s at index %d", row_data, new_child_index)
        item = _model.insert(receiver_row, new_child_index, row_data)

    return item or None


def select(receiver_path, device_number=None):
    assert _window
    assert receiver_path is not None
    if device_number is None:
        item = _receiver_row(receiver_path)
    else:
        item = _device_row(receiver_path, device_number)
    if item:
        selection = _tree.get_selection()
        selection.select_iter(item)
    else:
        logger.warning("select(%s, %s) failed to find an item", receiver_path, device_number)


def _hide(w, _ignore=None):
    assert w == _window
    # some window managers move the window to 0,0 after hide()
    # so try to remember the last position
    position = _window.get_position()
    _window.hide()
    _window.move(*position)
    return True


def popup(trigger=None, receiver_path=None, device_id=None):
    if receiver_path:
        select(receiver_path, device_id)
    _window.present()
    return True


def toggle(trigger=None):
    if _window.get_visible():
        _hide(_window)
    else:
        _window.present()


def _update_details(button):
    assert button
    visible = button.get_active()

    if visible:
        # _details._text.set_markup('<small>reading...</small>')

        def _details_items(device, read_all=False):
            # If read_all is False, only return stuff that is ~100% already
            # cached, and involves no HID++ calls.

            yield _("Path"), device.path
            if device.kind is None:
                yield _("USB ID"), f"{LOGITECH_VENDOR_ID:04x}:" + device.product_id

                if read_all:
                    yield _("Serial"), device.serial
                else:
                    yield _("Serial"), "..."

            else:
                # yield ('Codename', device.codename)
                yield _("Index"), device.number
                if device.wpid:
                    yield _("Wireless PID"), device.wpid
                if device.product_id:
                    yield _("Product ID"), f"{LOGITECH_VENDOR_ID:04x}:" + device.product_id
                hid_version = device.protocol
                yield _("Protocol"), f"HID++ {hid_version:1.1f}" if hid_version else _("Unknown")
                if read_all and device.polling_rate:
                    yield _("Polling rate"), device.polling_rate

                if read_all or not device.online:
                    yield _("Serial"), device.serial
                else:
                    yield _("Serial"), "..."
                if read_all and device.unitId and device.unitId != device.serial:
                    yield _("Unit ID"), device.unitId

            if read_all:
                if device.firmware:
                    for fw in list(device.firmware):
                        yield "  " + _(str(fw.kind)), (fw.name + " " + fw.version).strip()
            elif device.kind is None or device.online:
                yield f"  {_('Firmware')}", "..."

            flag_bits = device.notification_flags
            if flag_bits is not None:
                flag_names = hidpp10_constants.flags_to_str(flag_bits, fallback=f"({_('none')})")
                yield _("Notifications"), flag_names

        def _set_details(text):
            _details._text.set_markup(text)

        def _make_text(items):
            text = "\n".join("%-13s: %s" % (name, value) for name, value in items)
            return f"<small><tt>{text}</tt></small>"

        def _displayable_items(items):
            for name, value in items:
                value = GLib.markup_escape_text(str(value).replace("\x00", "")).strip()
                if value:
                    yield name, value

        def _read_slow(device):
            items = _details_items(selected_device, True)
            items = _displayable_items(items)
            text = _make_text(items)
            if device == _details._current_device:
                GLib.idle_add(_set_details, text)

        selected_device = _find_selected_device()
        assert selected_device
        _details._current_device = selected_device

        read_all = not (selected_device.kind is None or selected_device.online)
        items = _details_items(selected_device, read_all)
        _set_details(_make_text(items))

        if read_all:
            _details._current_device = None
        else:
            ui_async(_read_slow, selected_device)

    _details.set_visible(visible)


def _update_receiver_panel(receiver, panel, buttons, full=False):
    assert receiver

    devices_count = len(receiver)

    paired_text = (
        _(_("No device paired."))
        if devices_count == 0
        else ngettext("%(count)s paired device.", "%(count)s paired devices.", devices_count) % {"count": devices_count}
    )

    if receiver.max_devices > 0:
        paired_text += (
            "\n\n<small>%s</small>"
            % ngettext(
                "Up to %(max_count)s device can be paired to this receiver.",
                "Up to %(max_count)s devices can be paired to this receiver.",
                receiver.max_devices,
            )
            % {"max_count": receiver.max_devices}
        )
    elif devices_count > 0:
        paired_text += f"\n\n<small>{_('Only one device can be paired to this receiver.')}</small>"
    pairings = receiver.remaining_pairings()
    if pairings is not None and pairings >= 0:
        paired_text += "\n<small>%s</small>" % (
            ngettext("This receiver has %d pairing remaining.", "This receiver has %d pairings remaining.", pairings)
            % pairings
        )

    panel._count.set_markup(paired_text)

    is_pairing = receiver.pairing.lock_open
    if is_pairing:
        panel._scanning.set_visible(True)
        if not panel._spinner.get_visible():
            panel._spinner.start()
        panel._spinner.set_visible(True)
    else:
        panel._scanning.set_visible(False)
        if panel._spinner.get_visible():
            panel._spinner.stop()
        panel._spinner.set_visible(False)

    panel.set_visible(True)

    # b._insecure.set_visible(False)
    buttons._unpair.set_visible(False)

    if (
        not is_pairing
        and (receiver.remaining_pairings() is None or receiver.remaining_pairings() != 0)
        and (receiver.re_pairs or devices_count < receiver.max_devices)
    ):
        buttons._pair.set_sensitive(True)
    else:
        buttons._pair.set_sensitive(False)

    buttons._pair.set_visible(True)


def _update_device_panel(device, panel, buttons, full=False):
    assert device
    is_online = bool(device.online)
    panel.set_sensitive(is_online)

    if device.battery_info is None or device.battery_info.level is None:
        device.read_battery()

    battery_level = device.battery_info.level if device.battery_info is not None else None
    battery_voltage = device.battery_info.voltage if device.battery_info is not None else None
    if battery_level is None and battery_voltage is None:
        panel._battery.set_visible(False)
    else:
        panel._battery.set_visible(True)
        battery_next_level = device.battery_info.next_level
        charging = device.battery_info.charging() if device.battery_info is not None else None
        icon_name = icons.battery(battery_level, charging)
        panel._battery._icon.set_from_icon_name(icon_name, _INFO_ICON_SIZE)
        panel._battery._icon.set_sensitive(True)
        panel._battery._text.set_sensitive(is_online)

        if battery_voltage is not None:
            panel._battery._label.set_text(_("Battery Voltage"))
            text = f"{int(battery_voltage)}mV"
            tooltip_text = _("Voltage reported by battery")
        else:
            panel._battery._label.set_text(_("Battery Level"))
            text = ""
            tooltip_text = _("Approximate level reported by battery")
        if battery_voltage is not None and battery_level is not None:
            text += ", "
        if battery_level is not None:
            text += _(str(battery_level)) if isinstance(battery_level, NamedInt) else f"{int(battery_level)}%"
        if battery_next_level is not None and not charging:
            if isinstance(battery_next_level, NamedInt):
                text += "<small> (" + _("next reported ") + _(str(battery_next_level)) + ")</small>"
            else:
                text += "<small> (" + _("next reported ") + f"{int(battery_next_level)}%" + ")</small>"
            tooltip_text = tooltip_text + _(" and next level to be reported.")
        if is_online:
            if charging:
                text += f" <small>({_('charging')})</small>"
        else:
            text += f" <small>({_('last known')})</small>"

        panel._battery._text.set_markup(text)
        panel._battery.set_tooltip_text(tooltip_text)

    if device.link_encrypted is None:
        panel._secure.set_visible(False)
    elif is_online:
        panel._secure.set_visible(True)
        panel._secure._icon.set_visible(True)
        if device.link_encrypted is True:
            panel._secure._text.set_text(_("encrypted"))
            panel._secure._icon.set_from_icon_name("security-high", _INFO_ICON_SIZE)
            panel._secure.set_tooltip_text(_("The wireless link between this device and its receiver is encrypted."))
        else:
            panel._secure._text.set_text(_("not encrypted"))
            panel._secure._icon.set_from_icon_name("security-low", _INFO_ICON_SIZE)
            panel._secure.set_tooltip_text(
                _(
                    "The wireless link between this device and its receiver is not encrypted.\n"
                    "This is a security issue for pointing devices, and a major security issue for text-input devices."
                )
            )
    else:
        panel._secure.set_visible(True)
        panel._secure._icon.set_visible(False)
        panel._secure._text.set_markup(f"<small>{_('offline')}</small>")
        panel._secure.set_tooltip_text("")

    if is_online:
        light_level = device.battery_info.light_level if device.battery_info is not None else None
        if light_level is None:
            panel._lux.set_visible(False)
        else:
            panel._lux._icon.set_from_icon_name(icons.lux(light_level), _INFO_ICON_SIZE)
            panel._lux._text.set_text(_("%(light_level)d lux") % {"light_level": light_level})
            panel._lux.set_visible(True)
    else:
        panel._lux.set_visible(False)

    buttons._pair.set_visible(False)
    buttons._unpair.set_sensitive(device.receiver.may_unpair if device.receiver else False)
    buttons._unpair.set_visible(True)

    panel.set_visible(True)

    if full:
        config_panel.update(device, is_online)


def _update_info_panel(device, full=False):
    if device is None:
        # no selected device, show the 'empty' panel
        _details.set_visible(False)
        _info.set_visible(False)
        _empty.set_visible(True)
        return

    # a receiver must be valid
    # a device must be paired
    assert device

    _info._title.set_markup(f"<b>{device.name}</b>")
    icon_name = icons.device_icon_name(device.name, device.kind)
    _info._icon.set_from_icon_name(icon_name, _DEVICE_ICON_SIZE)

    if device.kind is None:
        _info._device.set_visible(False)
        _info._icon.set_sensitive(True)
        _info._title.set_sensitive(True)
        _update_receiver_panel(device, _info._receiver, _info._buttons, full)
    else:
        _info._receiver.set_visible(False)
        is_online = bool(device.online)
        _info._icon.set_sensitive(is_online)
        _info._title.set_sensitive(is_online)
        _update_device_panel(device, _info._device, _info._buttons, full)

    _empty.set_visible(False)
    _info.set_visible(True)

    if full:
        _update_details(_info._buttons._details)


# window layout:
#  +--------------------------------+
#  |  tree      | receiver  | empty |
#  |            | or device |       |
#  |------------| status    |       |
#  | details    |           |       |
#  |--------------------------------|
#  | (about)                        |
#  +--------------------------------|
# either the status or empty panel is visible at any point
# the details panel can be toggle on/off

_model = None
_tree = None
_details = None
_info = None
_empty = None
_window = None


def init(show_window, hide_on_close):
    Gtk.Window.set_default_icon_name(NAME.lower())

    global _model, _tree, _details, _info, _empty, _window
    _model = Gtk.TreeStore(*_COLUMN_TYPES)
    _tree = _create_tree(_model)
    _details = _create_details_panel()
    _info = _create_info_panel()
    _empty = _create_empty_panel()
    _window = _create(_hide if hide_on_close else destroy)
    if show_window:
        _window.present()


def destroy(_ignore1=None, _ignore2=None):
    global _model, _tree, _details, _info, _empty, _window
    w, _window = _window, None
    w.destroy()
    w = None
    config_panel.destroy()

    _empty = None
    _info = None
    _details = None
    _tree = None
    _model = None


def update(device, need_popup=False, refresh=False):
    if _window is None:
        return

    assert device is not None

    if need_popup:
        popup()

    selected_device_id = _find_selected_device_id()

    if device.kind is None:  # receiver
        # receiver
        is_alive = bool(device)
        item = _receiver_row(device.path, device if is_alive else None)

        if is_alive and item:
            was_pairing = bool(_model.get_value(item, Column.STATUS_ICON))
            is_pairing = (not device.isDevice) and bool(device.pairing.lock_open)
            _model.set_value(item, Column.STATUS_ICON, "network-wireless" if is_pairing else _CAN_SET_ROW_NONE)

            if selected_device_id == (device.path, 0):
                full_update = need_popup or was_pairing != is_pairing
                _update_info_panel(device, full=full_update)

        elif item:
            if _TREE_SEPATATOR:
                separator = _model.iter_next(item)
                if separator:
                    _model.remove(separator)
            _model.remove(item)

    else:
        path = device.receiver.path if device.receiver is not None else device.path
        assert device.number is not None and device.number >= 0, f"invalid device number{str(device.number)}"
        item = _device_row(path, device.number, device if bool(device) else None)

        if bool(device) and item:
            update_device(device, item, selected_device_id, need_popup, full=refresh)
        elif item:
            _model.remove(item)
            config_panel.clean(device)

    # make sure all rows are visible
    _tree.expand_all()


def update_device(device, item, selected_device_id, need_popup, full=False):
    was_online = _model.get_value(item, Column.ACTIVE)
    is_online = bool(device.online)
    _model.set_value(item, Column.ACTIVE, is_online)

    battery_level = device.battery_info.level if device.battery_info is not None else None
    battery_voltage = device.battery_info.voltage if device.battery_info is not None else None
    if battery_level is None:
        _model.set_value(item, Column.STATUS_TEXT, _CAN_SET_ROW_NONE)
        _model.set_value(item, Column.STATUS_ICON, _CAN_SET_ROW_NONE)
    else:
        if battery_voltage is not None and False:  # Use levels instead of voltage here
            status_text = f"{int(battery_voltage)}mV"
        elif isinstance(battery_level, NamedInt):
            status_text = _(str(battery_level))
        else:
            status_text = f"{int(battery_level)}%"
        _model.set_value(item, Column.STATUS_TEXT, status_text)

        charging = device.battery_info.charging() if device.battery_info is not None else None
        icon_name = icons.battery(battery_level, charging)
        _model.set_value(item, Column.STATUS_ICON, icon_name)

    _model.set_value(item, Column.NAME, device.codename)

    if selected_device_id is None or need_popup:
        select(device.receiver.path if device.receiver else device.path, device.number)
    elif selected_device_id == (device.receiver.path if device.receiver else device.path, device.number):
        full_update = full or was_online != is_online
        _update_info_panel(device, full=full_update)


def find_device(serial):
    assert serial, "need serial number or unit ID to find a device"
    result = None

    def check(_store, _treepath, row):
        nonlocal result
        device = _model.get_value(row, Column.DEVICE)
        if device and device.kind and (device.unitId == serial or device.serial == serial):
            result = device
            return True

    _model.foreach(check)
    return result



## Copyright (C) 2012-2013  Daniel Pavel
## Copyright (C) 2014-2024  Solaar Contributors https://pwr-solaar.github.io/Solaar/
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
from __future__ import annotations

import logging

from typing import Callable

logger = logging.getLogger(__name__)

try:
    import dbus

    from dbus.mainloop.glib import DBusGMainLoop  # integration into the main GLib loop

    DBusGMainLoop(set_as_default=True)
    bus = dbus.SystemBus()
    assert bus

except Exception:
    # Either the dbus library is not available or the system dbus is not running
    logger.warning("failed to set up dbus")
    bus = None


_suspend_callback = None
_resume_callback = None


def _suspend_or_resume(suspend):
    if suspend and _suspend_callback:
        _suspend_callback()
    if not suspend and _resume_callback:
        _resume_callback()


_LOGIND_PATH = "/org/freedesktop/login1"
_LOGIND_INTERFACE = "org.freedesktop.login1.Manager"


def watch_suspend_resume(
    on_resume_callback: Callable[[], None] | None = None,
    on_suspend_callback: Callable[[], None] | None = None,
):
    """Register callback for suspend/resume events.
    They are called only if the system DBus is running, and the Login daemon is available."""
    global _resume_callback, _suspend_callback
    _suspend_callback = on_suspend_callback
    _resume_callback = on_resume_callback
    if bus is not None and on_resume_callback is not None or on_suspend_callback is not None:
        bus.add_signal_receiver(
            _suspend_or_resume,
            "PrepareForSleep",
            dbus_interface=_LOGIND_INTERFACE,
            path=_LOGIND_PATH,
        )
    logger.info("connected to system dbus, watching for suspend/resume events")


_BLUETOOTH_PATH_PREFIX = "/org/bluez/hci0/dev_"
_BLUETOOTH_INTERFACE = "org.freedesktop.DBus.Properties"

_bluetooth_callbacks = {}


def watch_bluez_connect(serial, callback=None):
    if _bluetooth_callbacks.get(serial):
        _bluetooth_callbacks.get(serial).remove()
    path = _BLUETOOTH_PATH_PREFIX + serial.replace(":", "_").upper()
    if bus is not None and callback is not None:
        _bluetooth_callbacks[serial] = bus.add_signal_receiver(
            callback, "PropertiesChanged", path=path, dbus_interface=_BLUETOOTH_INTERFACE
        )



## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from time import time

from logitech_receiver import base
from logitech_receiver import hidpp10
from logitech_receiver import hidpp10_constants
from logitech_receiver import notifications

_hidpp10 = hidpp10.Hidpp10()


def run(receivers, args, find_receiver, _ignore):
    assert receivers

    if args.receiver:
        receiver_name = args.receiver.lower()
        receiver = find_receiver(receivers, receiver_name)
        if not receiver:
            raise Exception(f"no receiver found matching '{receiver_name}'")
    else:
        receiver = receivers[0]

    assert receiver

    # check if it's necessary to set the notification flags
    old_notification_flags = _hidpp10.get_notification_flags(receiver)
    if not (old_notification_flags & hidpp10_constants.NotificationFlag.WIRELESS):
        _hidpp10.set_notification_flags(receiver, old_notification_flags | hidpp10_constants.NotificationFlag.WIRELESS)

    # get all current devices
    known_devices = [dev.number for dev in receiver]

    class _HandleWithNotificationHook(int):
        def notifications_hook(self, n):
            nonlocal known_devices
            assert n
            if n.devnumber == 0xFF:
                notifications.process(receiver, n)
            elif n.sub_id == 0x41 and len(n.data) == base.SHORT_MESSAGE_SIZE - 4:
                kd, known_devices = known_devices, None  # only process one connection notification
                if kd is not None:
                    if n.devnumber not in kd:
                        receiver.pairing.new_device = receiver.register_new_device(n.devnumber, n)
                    elif receiver.re_pairs:
                        del receiver[n.devnumber]  # get rid of information on device re-paired away
                        receiver.pairing.new_device = receiver.register_new_device(n.devnumber, n)

    timeout = 30  # seconds
    receiver.handle = _HandleWithNotificationHook(receiver.handle)

    if receiver.receiver_kind == "bolt":  # Bolt receivers require authentication to pair a device
        receiver.discover(timeout=timeout)
        print("Bolt Pairing: long-press the pairing key or button on your device (timing out in", timeout, "seconds).")
        pairing_start = time()
        patience = 5  # the discovering notification may come slightly later, so be patient
        while receiver.pairing.discovering or time() - pairing_start < patience:
            if receiver.pairing.device_address and receiver.pairing.device_authentication and receiver.pairing.device_name:
                break
            n = base.read(receiver.handle)
            n = base.make_notification(*n) if n else None
            if n:
                receiver.handle.notifications_hook(n)
        address = receiver.pairing.device_address
        name = receiver.pairing.device_name
        authentication = receiver.pairing.device_authentication
        kind = receiver.pairing.device_kind
        print(f"Bolt Pairing: discovered {name}")
        receiver.pair_device(
            address=address,
            authentication=authentication,
            entropy=20 if kind == hidpp10_constants.DEVICE_KIND.keyboard else 10,
        )
        pairing_start = time()
        patience = 5  # the discovering notification may come slightly later, so be patient
        while receiver.pairing.lock_open or time() - pairing_start < patience:
            if receiver.pairing.device_passkey:
                break
            n = base.read(receiver.handle)
            n = base.make_notification(*n) if n else None
            if n:
                receiver.handle.notifications_hook(n)
        if authentication & 0x01:
            print(f"Bolt Pairing: type passkey {receiver.pairing.device_passkey} and then press the enter key")
        else:
            passkey = f"{int(receiver.pairing.device_passkey):010b}"
            passkey = ", ".join(["right" if bit == "1" else "left" for bit in passkey])
            print(f"Bolt Pairing: press {passkey}")
            print("and then press left and right buttons simultaneously")
        while receiver.pairing.lock_open:
            n = base.read(receiver.handle)
            n = base.make_notification(*n) if n else None
            if n:
                receiver.handle.notifications_hook(n)

    else:
        receiver.set_lock(False, timeout=timeout)
        print("Pairing: Turn your device on or press, hold, and release")
        print("a channel button or the channel switch button.")
        print("Timing out in", timeout, "seconds.")
        pairing_start = time()
        patience = 5  # the lock-open notification may come slightly later, wait for it a bit
        while receiver.pairing.lock_open or time() - pairing_start < patience:
            n = base.read(receiver.handle)
            if n:
                n = base.make_notification(*n)
                if n:
                    receiver.handle.notifications_hook(n)

    if not (old_notification_flags & hidpp10_constants.NotificationFlag.WIRELESS):
        # only clear the flags if they weren't set before, otherwise a
        # concurrently running Solaar app might stop working properly
        _hidpp10.set_notification_flags(receiver, old_notification_flags)

    if receiver.pairing.new_device:
        dev = receiver.pairing.new_device
        print(f"Paired device {int(dev.number)}: {dev.name} ({dev.codename}) [{dev.wpid}:{dev.serial}]")
    else:
        error = receiver.pairing.error
        if error:
            raise Exception(f"pairing failed: {error}")
        else:
            print("Paired device")  # this is better than an error



## Copyright (C) 2012-2013  Daniel Pavel
## Copyright (C) 2014-2024  Solaar Contributors https://pwr-solaar.github.io/Solaar/
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from .common import KwException

"""Exceptions that may be raised by this API."""


class NoReceiver(KwException):
    """Raised when trying to talk through a previously open handle, when the
    receiver is no longer available. Should only happen if the receiver is
    physically disconnected from the machine, or its kernel driver module is
    unloaded."""

    pass


class NoSuchDevice(KwException):
    """Raised when trying to reach a device number not paired to the receiver."""

    pass


class DeviceUnreachable(KwException):
    """Raised when a request is made to an unreachable (turned off) device."""

    pass


class FeatureNotSupported(KwException):
    """Raised when trying to request a feature not supported by the device."""

    pass


class FeatureCallError(KwException):
    """Raised if the device replied to a feature call with an error."""

    pass



## Copyright (C) 2012-2013  Daniel Pavel
## Copyright (C) 2014-2024  Solaar Contributors https://pwr-solaar.github.io/Solaar/
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import logging
import queue
import threading

from . import base
from . import exceptions

logger = logging.getLogger(__name__)


class _ThreadedHandle:
    """A thread-local wrapper with different open handles for each thread.
    Closing a ThreadedHandle will close all handles.
    """

    __slots__ = ("path", "_local", "_handles", "_listener")

    def __init__(self, listener, path, handle):
        assert listener is not None
        assert path is not None
        assert handle is not None
        assert isinstance(handle, int)

        self._listener = listener
        self.path = path
        self._local = threading.local()
        # take over the current handle for the thread doing the replacement
        self._local.handle = handle
        self._handles = [handle]

    def _open(self):
        handle = base.open_path(self.path)
        if handle is None:
            logger.error("%r failed to open new handle", self)
        else:
            # if logger.isEnabledFor(logging.DEBUG):
            #     logger.debug("%r opened new handle %d", self, handle)
            self._local.handle = handle
            self._handles.append(handle)
            return handle

    def close(self):
        if self._local:
            self._local = None
            handles, self._handles = self._handles, []
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("%r closing %s", self, handles)
            for h in handles:
                base.close(h)

    @property
    def notifications_hook(self):
        if self._listener:
            assert isinstance(self._listener, threading.Thread)
            if threading.current_thread() == self._listener:
                return self._listener._notifications_hook

    def __del__(self):
        self._listener = None
        self.close()

    def __index__(self):
        if self._local:
            try:
                return self._local.handle
            except Exception:
                return self._open()
        else:
            return -1

    __int__ = __index__

    def __str__(self):
        if self._local:
            return str(int(self))

    def __repr__(self):
        return f"<_ThreadedHandle({self.path})>"

    def __bool__(self):
        return bool(self._local)

    __nonzero__ = __bool__


# How long to wait during a read for the next packet, in seconds.
# Ideally this should be rather long (10s ?), but the read is blocking and this means that when the thread
# is signalled to stop, it would take a while for it to acknowledge it.
# Forcibly closing the file handle on another thread does _not_ interrupt the read on Linux systems.
_EVENT_READ_TIMEOUT = 1.0  # in seconds


class EventsListener(threading.Thread):
    """Listener thread for notifications from the Unifying Receiver.
    Incoming packets will be passed to the callback function in sequence.
    """

    def __init__(self, receiver, notifications_callback):
        try:
            path_name = receiver.path.split("/")[2]
        except IndexError:
            path_name = receiver.path
        super().__init__(name=f"{self.__class__.__name__}:{path_name}")
        self.daemon = True
        self._active = False
        self.receiver = receiver
        self._queued_notifications = queue.Queue(16)
        self._notifications_callback = notifications_callback

    def run(self):
        self._active = True
        # replace the handle with a threaded one
        self.receiver.handle = _ThreadedHandle(self, self.receiver.path, self.receiver.handle)
        if logger.isEnabledFor(logging.INFO):
            logger.info("started with %s (%d)", self.receiver, int(self.receiver.handle))
        self.has_started()

        if self.receiver.isDevice:  # ping (wired or BT) devices to see if they are really online
            if self.receiver.ping():
                self.receiver.changed(active=True, reason="initialization")

        while self._active:
            if self._queued_notifications.empty():
                try:
                    n = base.read(self.receiver.handle, _EVENT_READ_TIMEOUT)
                except exceptions.NoReceiver:
                    logger.warning("%s disconnected", self.receiver.name)
                    self.receiver.close()
                    break
                if n:
                    n = base.make_notification(*n)
            else:
                n = self._queued_notifications.get()  # deliver any queued notifications
            if n:
                try:
                    self._notifications_callback(n)
                except Exception:
                    logger.exception("processing %s", n)

        del self._queued_notifications
        self.has_stopped()

    def stop(self):
        """Tells the listener to stop as soon as possible."""
        self._active = False

    def has_started(self):
        """Called right after the thread has started, and before it starts
        reading notification packets."""
        pass

    def has_stopped(self):
        """Called right before the thread stops."""
        pass

    def _notifications_hook(self, n):
        # Only consider unhandled notifications that were sent from this thread,
        # i.e. triggered by a callback handling a previous notification.
        assert threading.current_thread() == self
        if self._active:  # and threading.current_thread() == self:
            # if logger.isEnabledFor(logging.DEBUG):
            #     logger.debug("queueing unhandled %s", n)
            if not self._queued_notifications.full():
                self._queued_notifications.put(n)

    def __bool__(self):
        return bool(self._active and self.receiver)

    __nonzero__ = __bool__
